"""The model decorator."""
from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Dict
from typing import List

from glotaran.model.attribute import model_attribute_typed
from glotaran.model.dataset_descriptor import DatasetDescriptor
from glotaran.model.megacomplex import Megacomplex
from glotaran.model.util import wrap_func_as_method
from glotaran.model.weight import Weight
from glotaran.plugin_system.model_registration import register_model

if TYPE_CHECKING:
    from typing import Any
    from typing import Callable
    from typing import Tuple
    from typing import Type
    from typing import Union

    import numpy as np
    import xarray as xr

    from glotaran.analysis.problem import Problem
    from glotaran.model.base_model import Model
    from glotaran.parameter import ParameterGroup

    MegacomplexMatrixFunction = Callable[
        [Type[object], Type[Model], Type[DatasetDescriptor], dict[str, int], Any],
        Tuple[List[str], np.ndarray],
    ]
    """A `MatrixFunction` calculates the matrix for a model."""

    GlobalMatrixFunction = Callable[
        [Type[DatasetDescriptor], np.ndarray], Tuple[List[str], np.ndarray]
    ]
    """A `GlobalMatrixFunction` calculates the global matrix for a model."""

    ConstrainMatrixFunction = Callable[
        [Type[Model], ParameterGroup, List[str], np.ndarray, float],
        Tuple[List[str], np.ndarray],
    ]
    """A `ConstrainMatrixFunction` applies constraints on a matrix."""

    RetrieveClpFunction = Callable[
        [
            Type[Model],
            ParameterGroup,
            Dict[str, Union[List[str], List[List[str]]]],
            Dict[str, Union[List[str], List[List[str]]]],
            Dict[str, List[np.ndarray]],
            Dict[str, xr.Dataset],
        ],
        Dict[str, List[np.ndarray]],
    ]
    """A `RetrieveClpFunction` retrieves the full set of clp from a reduced set."""

    FinalizeFunction = Callable[[Problem, Dict[str, xr.Dataset]], None]
    """A `FinalizeFunction` gets called after optimization."""

    PenaltyFunction = Callable[
        [
            Type[Model],
            ParameterGroup,
            Dict[str, Union[List[str], List[List[str]]]],
            Dict[str, List[np.ndarray]],
            Dict[str, Union[np.ndarray, List[np.ndarray]]],
            Dict[str, xr.Dataset],
            float,
        ],
        np.ndarray,
    ]
    """A `PenaltyFunction` calculates additional penalties for the optimization."""


def model(
    model_type: str,
    attributes: dict[str, Any] = None,
    dataset_type: type[DatasetDescriptor] = DatasetDescriptor,
    default_megacomplex_type: str = None,
    megacomplex_types: dict[str, Megacomplex] | type[Megacomplex] = None,
    global_matrix: GlobalMatrixFunction = None,
    model_dimension: str = None,
    global_dimension: str = None,
    has_matrix_constraints_function: Callable[[type[Model]], bool] = None,
    constrain_matrix_function: ConstrainMatrixFunction = None,
    retrieve_clp_function: RetrieveClpFunction = None,
    has_additional_penalty_function: Callable[[type[Model]], bool] = None,
    additional_penalty_function: PenaltyFunction = None,
    finalize_data_function: FinalizeFunction = None,
    grouped: bool | Callable[[type[Model]], bool] = False,
    index_dependent: bool | Callable[[type[Model]], bool] = False,
) -> Callable[[type[Model]], type[Model]]:
    """The `@model` decorator is intended to be used on subclasses of :class:`glotaran.model.Model`.
    It creates properties for the given attributes as well as functions to add access them. Also it
    adds the functions (e.g. for `matrix`) to the model ensures they are added wrapped in a correct
    way.

    Parameters
    ----------
    model_type : str
        Human readable string used by the parser to identify the correct model.
    attributes : Dict[str, Any], optional
        A dictionary of attribute names and types. All types must be decorated with the
        :func:`glotaran.model.model_attribute` decorator, by default None.
    dataset_type : Type[DatasetDescriptor], optional
        A subclass of :class:`DatasetDescriptor`, by default DatasetDescriptor
    megacomplex_type : Any, optional
        A class for the model megacomplexes. The class must be decorated with the
        :func:`glotaran.model.model_attribute` decorator, by default None
    matrix : Union[MatrixFunction, IndexDependentMatrixFunction], optional
        A function to calculate the matrix for the model, by default None
    global_matrix : GlobalMatrixFunction, optional
        A function to calculate the global matrix for the model, by default None
    model_dimension : str, optional
        The name of model matrix row dimension, by default None
    global_dimension : str, optional
        The name of model global matrix row dimension, by default None
    has_matrix_constraints_function : Callable[[Type[Model]], bool], optional
        True if the model as a constrain_matrix_function set, by default None
    constrain_matrix_function : ConstrainMatrixFunction, optional
        A function to constrain the global matrix for the model, by default None
    retrieve_clp_function : RetrieveClpFunction, optional
        A function to retrieve the full clp from the reduced, by default None
    has_additional_penalty_function : Callable[[Type[Model]], bool], optional
        True if model has a additional_penalty_function set, by default None
    additional_penalty_function : PenaltyFunction, optional
        A function to calculate additional penalties when optimizing the model, by default None
    finalize_data_function : FinalizeFunction, optional
        A function to finalize data after optimization, by default None
    grouped : Union[bool, Callable[[Type[Model]], bool]], optional
        True if model described a grouped problem, by default False
    index_dependent : Union[bool, Callable[[Type[Model]], bool]], optional
        True if model described a index dependent problem, by default False

    Returns
    -------
    Callable
        Returns a decorated model function

    Raises
    ------
    ValueError
        If model implements meth:`has_matrix_constraints_function` but not
        meth:`constrain_matrix_function` and meth:`retrieve_clp_function`
    ValueError
        If model implements meth:`has_additional_penalty_function` but not
        meth:`additional_penalty_function`
    """

    def decorator(cls):

        setattr(cls, "_model_type", model_type)
        setattr(cls, "finalize_data", finalize_data_function)

        _set_constraints_functions(
            cls, has_matrix_constraints_function, constrain_matrix_function, retrieve_clp_function
        )

        _set_additional_penalty_functions(
            cls, has_additional_penalty_function, additional_penalty_function
        )

        _set_grouped_and_indexdependent(cls, grouped, index_dependent)

        _set_dimensions(cls, model_type, model_dimension, global_dimension)

        if global_matrix:
            g_mat = wrap_func_as_method(cls, name="global_matrix")(global_matrix)
            g_mat = staticmethod(g_mat)
            setattr(cls, "global_matrix", g_mat)
        else:
            setattr(cls, "global_matrix", None)

        if not hasattr(cls, "_glotaran_model_attributes"):
            setattr(cls, "_glotaran_model_attributes", {})
        else:
            setattr(
                cls,
                "_glotaran_model_attributes",
                getattr(cls, "_glotaran_model_attributes").copy(),
            )

        megacomplex_cls = _set_megacomplexes(
            cls, model_type, default_megacomplex_type, megacomplex_types
        )

        # We add the standard attributes here.
        if not issubclass(dataset_type, DatasetDescriptor):
            raise ValueError(
                f"Dataset descriptor of model {model_type} is not a subclass of DatasetDescriptor"
            )
        attributes["dataset"] = dataset_type
        attributes["megacomplex"] = megacomplex_cls
        attributes["weights"] = Weight

        # Set annotations and methods for attributes
        for attr_name, attr_type in attributes.items():

            # store for internal lookups
            getattr(cls, "_glotaran_model_attributes")[attr_name] = None

            # create and attach the property to class
            attr_prop = _create_property_for_attribute(cls, attr_name, attr_type)
            setattr(cls, attr_name, attr_prop)

            # properties with labels are implemented as dicts, whereas properties
            # without as arrays. Thus the need different setters.
            if getattr(attr_type, "_glotaran_has_label"):
                get_item = _create_get_func(cls, attr_name, attr_type)
                setattr(cls, get_item.__name__, get_item)
                set_item = _create_set_func(cls, attr_name, attr_type)
                setattr(cls, set_item.__name__, set_item)

            else:
                add_item = _create_add_func(cls, attr_name, attr_type)
                setattr(cls, add_item.__name__, add_item)

        init = _create_init_func(cls, attributes)
        setattr(cls, "__init__", init)

        register_model(model_type, cls)

        return cls

    return decorator


def _create_init_func(cls, attributes):
    @wrap_func_as_method(cls)
    def __init__(self):
        for attr_name, attr_item in attributes.items():
            if getattr(attr_item, "_glotaran_has_label"):
                setattr(self, f"_{attr_name}", {})
            else:
                setattr(self, f"_{attr_name}", [])
        super(cls, self).__init__()

    return __init__


def _create_add_func(cls, name, item_type):
    @wrap_func_as_method(cls, name=f"add_{name}", annotations={"item": item_type})
    def add_item(self, item: item_type):
        f"""Adds an `{item_type.__name__}` object.

        Parameters
        ----------
        item :
            The `{item_type.__name__}` item.
        """

        if not isinstance(item, item_type) and (
            not hasattr(item_type, "_glotaran_model_attribute_typed")
            or not isinstance(item, tuple(item_type._glotaran_model_attribute_types.values()))
        ):
            raise TypeError
        getattr(self, f"_{name}").append(item)

    return add_item


def _create_get_func(cls, name, item_type):
    @wrap_func_as_method(cls, name=f"get_{name}", annotations={"return": item_type})
    def get_item(self, label: str) -> item_type:
        f"""
        Returns the `{item_type.__name__}` object with the given label.

        Parameters
        ----------
        label :
            The label of the `{item_type.__name__}` object.
        """
        return getattr(self, f"_{name}")[label]

    return get_item


def _create_set_func(cls, name, item_type):
    @wrap_func_as_method(cls, name=f"set_{name}", annotations={"item": item_type})
    def set_item(self, label: str, item: item_type):
        f"""
        Sets the `{item_type.__name__}` object with the given label with to the item.

        Parameters
        ----------
        label :
            The label of the `{item_type.__name__}` object.
        item :
            The `{item_type.__name__}` item.
        """

        if (
            not isinstance(item, item_type)
            and (
                not hasattr(item_type, "_glotaran_model_attribute_typed")
                or not isinstance(item, tuple(item_type._glotaran_model_attribute_types.values()))
            )
            and not isinstance(item, Megacomplex)
        ):
            raise TypeError
        getattr(self, f"_{name}")[label] = item

    return set_item


def _set_megacomplexes(cls, model_type, default_megacomplex_type, megacomplex_types):
    @model_attribute_typed({})
    class MetaMegacomplex:
        """This class holds all Megacomplex types defined by a model."""

    if not isinstance(megacomplex_types, dict):
        megacomplex_types = {model_type: megacomplex_types}
    for name, megacomplex_type in megacomplex_types.items():
        if not issubclass(megacomplex_type, Megacomplex):
            raise TypeError(
                f"Megacomplex type {name}(megacomplex_type) is not a subclass of Megacomplex"
            )
        MetaMegacomplex.add_type(name, megacomplex_type)

    if default_megacomplex_type is None:
        default_megacomplex_type = next(iter(megacomplex_types.keys()))
    setattr(MetaMegacomplex, "_glotaran_model_attribute_default_type", default_megacomplex_type)
    return MetaMegacomplex


def _create_property_for_attribute(cls, name, attribute_type):

    return_type = (
        Dict[str, attribute_type]
        if hasattr(attribute_type, "_glotaran_has_label")
        else List[attribute_type]
    )

    doc_type = "dictionary" if hasattr(attribute_type, "_glotaran_has_label") else "list"

    @property
    @wrap_func_as_method(cls, name=f"{name}", annotations={"return": return_type})
    def attribute(self) -> return_type:
        f"""A {doc_type} containing {type.__name__}"""
        return getattr(self, f"_{name}")

    return attribute


def _set_constraints_functions(
    cls, has_matrix_constraints_function, constrain_matrix_function, retrieve_clp_function
):
    if has_matrix_constraints_function:
        if not constrain_matrix_function:
            raise ValueError(
                "Model implements `has_matrix_constraints_function` "
                "but not `constrain_matrix_function`"
            )
        if not retrieve_clp_function:
            raise ValueError(
                "Model implements `has_matrix_constraints_function` "
                "but not `retrieve_clp_function`"
            )
        has_c_mat = wrap_func_as_method(cls, name="has_matrix_constraints_function")(
            has_matrix_constraints_function
        )
        c_mat = wrap_func_as_method(cls, name="constrain_matrix_function")(
            constrain_matrix_function
        )
        r_clp = wrap_func_as_method(cls, name="retrieve_clp_function")(retrieve_clp_function)
        setattr(cls, "has_matrix_constraints_function", has_c_mat)
        setattr(cls, "constrain_matrix_function", c_mat)
        setattr(cls, "retrieve_clp_function", r_clp)
    else:
        setattr(cls, "has_matrix_constraints_function", None)
        setattr(cls, "constrain_matrix_function", None)
        setattr(cls, "retrieve_clp_function", None)


def _set_additional_penalty_functions(
    cls, has_additional_penalty_function, additional_penalty_function
):
    if has_additional_penalty_function:
        if not additional_penalty_function:
            raise ValueError(
                "Model implements `has_additional_penalty_function`"
                "but not `additional_penalty_function`"
            )
        has_pen = wrap_func_as_method(cls, name="has_additional_penalty_function")(
            has_additional_penalty_function
        )
        pen = wrap_func_as_method(cls, name="additional_penalty_function")(
            additional_penalty_function
        )
        setattr(cls, "additional_penalty_function", pen)
        setattr(cls, "has_additional_penalty_function", has_pen)
    else:
        setattr(cls, "has_additional_penalty_function", None)
        setattr(cls, "additional_penalty_function", None)


def _set_grouped_and_indexdependent(cls, grouped, index_dependent):
    setattr(
        cls,
        "grouped",
        grouped if callable(grouped) else lambda model: grouped,
    )

    setattr(
        cls,
        "index_dependent",
        index_dependent if callable(index_dependent) else lambda model: index_dependent,
    )


def _set_dimensions(cls, model_type, model_dimension, global_dimension):
    if model_dimension is None:
        raise ValueError(f"Model dimension not specified for model {model_type}")
    setattr(cls, "model_dimension", model_dimension)

    if global_dimension is None:
        raise ValueError(f"Global dimension not specified for model {model_type}")
    setattr(cls, "global_dimension", global_dimension)



from __future__ import annotations

from typing import Any

import numpy as np

from glotaran.builtin.models.kinetic_spectrum.kinetic_spectrum_model import KineticSpectrumModel
from glotaran.model import model_attribute
from glotaran.parameter import Parameter
from glotaran.parameter import ParameterGroup

class EqualAreaPenalty:
    @property
    def compartment(self) -> str: ...
    @property
    def interval(self) -> list[tuple[float, float]]: ...
    @property
    def target(self) -> str: ...
    @property
    def parameter(self) -> Parameter: ...
    @property
    def weight(self) -> str: ...
    def applies(self, index: Any) -> bool: ...

def has_spectral_penalties(model: KineticSpectrumModel) -> bool: ...
def apply_spectral_penalties(
    model: KineticSpectrumModel,
    parameters: ParameterGroup,
    clp_labels: list[str] | list[list[str]],
    clps: np.ndarray,
    global_axis: np.ndarray,
) -> np.ndarray: ...



"""The parameter class."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING

import asteval
import numpy as np
from numpy.typing._array_like import _SupportsArray

if TYPE_CHECKING:
    from typing import Any

    from glotaran.parameter import ParameterGroup

RESERVED_LABELS: list[str] = list(asteval.make_symbol_table().keys()) + ["group"]


class Keys:
    """Keys for parameter options."""

    EXPR = "expr"
    MAX = "max"
    MIN = "min"
    NON_NEG = "non-negative"
    VARY = "vary"


class Parameter(_SupportsArray):
    """A parameter for optimization."""

    _find_parameter = re.compile(r"(\$[\w\d\.]+)")
    """A regexpression to find and replace parameter names in expressions."""
    _label_validator_regexp = re.compile(r"\W", flags=re.ASCII)
    """A regexpression to validate labels."""

    def __init__(
        self,
        label: str = None,
        full_label: str = None,
        expression: str = None,
        maximum: int | float = np.inf,
        minimum: int | float = -np.inf,
        non_negative: bool = False,
        value: float | int = np.nan,
        vary: bool = True,
    ):
        """Optimization Parameter supporting numpy array operations.

        Parameters
        ----------
        label : str, optional
            The label of the parameter., by default None
        full_label : str, optional
            The label of the parameter with its path in a parameter group prepended.
            , by default None
        expression : str, optional
            Expression to calculate the parameters value from,
            e.g. if used in relation to another parameter. , by default None
        maximum : int, optional
            Upper boundary for the parameter to be varied to., by default np.inf
        minimum : int, optional
            Lower boundary for the parameter to be varied to., by default -np.inf
        non_negative : bool, optional
            Whether the parameter should always be bigger than zero., by default False
        value : float, optional
            Value of the parameter, by default np.nan
        vary : bool, optional
            Whether the parameter should be changed during optimization or not.
            , by default True
        """

        self.label = label
        self.full_label = full_label or ""
        self.expression = expression
        self.maximum = maximum
        self.minimum = minimum
        self.non_negative = non_negative
        self.standard_error = 0.0
        self.value = value
        self.vary = vary

        self._transformed_expression: str | None = None

    @classmethod
    def valid_label(cls, label: str) -> bool:
        """Returns true if the `label` is valid string."""
        return cls._label_validator_regexp.search(label) is None and label not in RESERVED_LABELS

    @classmethod
    def from_list_or_value(
        cls,
        value: int | float | list,
        default_options: dict = None,
        label: str = None,
    ) -> Parameter:
        """Creates a parameter from a list or numeric value.

        Parameters
        ----------
        value :
            The list or numeric value.
        default_options :
            A dictionary of default options.
        label :
            The label of the parameter.
        """
        param = cls(label=label)
        options = None

        if not isinstance(value, list):
            param.value = value

        else:
            values = _sanatize_parameter_list(value)
            param.label = _retrieve_from_list_by_type(values, str, label)
            param.value = float(_retrieve_from_list_by_type(values, (int, float), 0))
            options = _retrieve_from_list_by_type(values, dict, None)

        if default_options:
            param._set_options_from_dict(default_options)

        if options:
            param._set_options_from_dict(options)
        return param

    def set_from_group(self, group: ParameterGroup):
        """Sets all values of the parameter to the values of the corresponding parameter in the group.

        Notes
        -----

        For internal use.

        Parameters
        ----------
        group :
            The :class:`glotaran.parameter.ParameterGroup`.
        """

        p = group.get(self.full_label)
        self.expression = p.expression
        self.maximum = p.maximum
        self.minimum = p.minimum
        self.non_negative = p.non_negative
        self.standard_error = p.standard_error
        self.value = p.value
        self.vary = p.vary

    def _set_options_from_dict(self, options: dict):
        if Keys.EXPR in options:
            self.expression = options[Keys.EXPR]
        if Keys.NON_NEG in options:
            self.non_negative = options[Keys.NON_NEG]
        if Keys.MAX in options:
            self.maximum = options[Keys.MAX]
        if Keys.MIN in options:
            self.minimum = options[Keys.MIN]
        if Keys.VARY in options:
            self.vary = options[Keys.VARY]

    @property
    def label(self) -> str | None:
        """Label of the parameter"""
        return self._label

    @label.setter
    def label(self, label: str | None):
        # ensure that label is str | None even if an int is passed
        label = None if label is None else str(label)
        if label is not None and not Parameter.valid_label(label):
            raise ValueError(f"'{label}' is not a valid group label.")
        self._label = label

    @property
    def full_label(self) -> str:
        """The label of the parameter with its path in a parameter group prepended."""
        return self._full_label

    @full_label.setter
    def full_label(self, full_label: str):
        self._full_label = str(full_label)

    @property
    def non_negative(self) -> bool:
        r"""Indicates if the parameter is non-negativ.

        If true, the parameter will be transformed with :math:`p' = \log{p}` and
        :math:`p = \exp{p'}`.

        Always `False` if `expression` is not `None`.
        """  # w605
        return self._non_negative if self.expression is None else False

    @non_negative.setter
    def non_negative(self, non_negative: bool):
        self._non_negative = non_negative

    @property
    def vary(self) -> bool:
        """Indicates if the parameter should be optimized.

        Always `False` if `expression` is not `None`.
        """
        return self._vary if self.expression is None else False

    @vary.setter
    def vary(self, vary: bool):
        self._vary = vary

    @property
    def maximum(self) -> float:
        """The upper bound of the parameter."""
        return self._maximum

    @maximum.setter
    def maximum(self, maximum: int | float):
        if not isinstance(maximum, float):
            try:
                maximum = float(maximum)
            except Exception:
                raise TypeError(
                    "Parameter maximum must be numeric."
                    + f"'{self.full_label}' has maximum '{maximum}' of type '{type(maximum)}'"
                )

        self._maximum = maximum

    @property
    def minimum(self) -> float:
        """The lower bound of the parameter."""
        return self._minimum

    @minimum.setter
    def minimum(self, minimum: int | float):
        if not isinstance(minimum, float):
            try:
                minimum = float(minimum)
            except Exception:
                raise TypeError(
                    "Parameter minimum must be numeric."
                    + f"'{self.full_label}' has minimum '{minimum}' of type '{type(minimum)}'"
                )

        self._minimum = minimum

    @property
    def expression(self) -> str | None:
        """Expression to calculate the parameters value from.

        This can used to set a relation to another parameter.
        """
        return self._expression

    @expression.setter
    def expression(self, expression: str | None):
        self._expression = expression
        self._transformed_expression = None

    @property
    def transformed_expression(self) -> str | None:
        """The expression of the parameter transformed for evaluation within a `ParameterGroup`."""
        if self.expression is not None and self._transformed_expression is None:
            self._transformed_expression = self.expression
            for match in Parameter._find_parameter.findall(self._transformed_expression):
                self._transformed_expression = self._transformed_expression.replace(
                    match, f"group.get('{match[1:]}').value"
                )
        return self._transformed_expression

    @property
    def standard_error(self) -> float:
        """The standard error of the optimized parameter."""
        return self._stderr

    @standard_error.setter
    def standard_error(self, standard_error: float):
        self._stderr = standard_error

    @property
    def value(self) -> float:
        """The value of the parameter"""
        return self._getval()

    @value.setter
    def value(self, value: int | float):
        if not isinstance(value, float) and value is not np.nan:
            try:
                value = float(value)
            except Exception:
                raise TypeError(
                    "Parameter value must be numeric."
                    + f"'{self.full_label}' has value '{value}' of type '{type(value)}'"
                )

        self._value = value

    def get_value_and_bounds_for_optimization(self) -> tuple[float, float, float]:
        """Gets the parameter value and bounds with expression and non-negative constraints
        applied."""
        value = self.value
        minimum = self.minimum
        maximum = self.maximum

        if self.non_negative:
            value = _log_value(value)
            minimum = _log_value(minimum)
            maximum = _log_value(maximum)

        return value, minimum, maximum

    def set_value_from_optimization(self, value: float):
        """Sets the value from an optimization result and reverses non-negative transformation."""
        self.value = np.exp(value) if self.non_negative else value

    def __getstate__(self):
        """Get state for pickle."""
        return (
            self.label,
            self.full_label,
            self.expression,
            self.maximum,
            self.minimum,
            self.non_negative,
            self.standard_error,
            self.value,
            self.vary,
        )

    def __setstate__(self, state):
        """Set state from pickle."""
        (
            self.label,
            self.full_label,
            self.expression,
            self.maximum,
            self.minimum,
            self.non_negative,
            self.standard_error,
            self.value,
            self.vary,
        ) = state

    def _getval(self) -> float:
        return self._value

    def __repr__(self):
        """Representation used by repl and tracebacks."""
        return (
            f"{type(self).__name__}(label={self.label!r}, value={self.value!r},"
            f" expression={self.expression!r}, vary={self.vary!r})"
        )

    def __array__(self):
        """array"""
        return np.array(float(self._getval()), dtype=float)

    def __str__(self):
        """Representation used by print and str."""
        return (
            f"__{self.label}__: _Value_: {self.value}, _StdErr_: {self.standard_error}, _Min_:"
            f" {self.minimum}, _Max_: {self.maximum}, _Vary_: {self.vary},"
            f" _Non-Negative_: {self.non_negative}, _Expr_: {self.expression}"
        )

    def __abs__(self):
        """abs"""
        return abs(self._getval())

    def __neg__(self):
        """neg"""
        return -self._getval()

    def __pos__(self):
        """positive"""
        return +self._getval()

    def __int__(self):
        """int"""
        return int(self._getval())

    def __float__(self):
        """float"""
        return float(self._getval())

    def __trunc__(self):
        """trunc"""
        return self._getval().__trunc__()

    def __add__(self, other):
        """+"""
        return self._getval() + other

    def __sub__(self, other):
        """-"""
        return self._getval() - other

    def __truediv__(self, other):
        """/"""
        return self._getval() / other

    def __floordiv__(self, other):
        """//"""
        return self._getval() // other

    def __divmod__(self, other):
        """divmod"""
        return divmod(self._getval(), other)

    def __mod__(self, other):
        """%"""
        return self._getval() % other

    def __mul__(self, other):
        """*"""
        return self._getval() * other

    def __pow__(self, other):
        """**"""
        return self._getval() ** other

    def __gt__(self, other):
        """>"""
        return self._getval() > other

    def __ge__(self, other):
        """>="""
        return self._getval() >= other

    def __le__(self, other):
        """<="""
        return self._getval() <= other

    def __lt__(self, other):
        """<"""
        return self._getval() < other

    def __eq__(self, other):
        """=="""
        return self._getval() == other

    def __ne__(self, other):
        """!="""
        return self._getval() != other

    def __radd__(self, other):
        """+ (right)"""
        return other + self._getval()

    def __rtruediv__(self, other):
        """/ (right)"""
        return other / self._getval()

    def __rdivmod__(self, other):
        """divmod (right)"""
        return divmod(other, self._getval())

    def __rfloordiv__(self, other):
        """// (right)"""
        return other // self._getval()

    def __rmod__(self, other):
        """% (right)"""
        return other % self._getval()

    def __rmul__(self, other):
        """* (right)"""
        return other * self._getval()

    def __rpow__(self, other):
        """** (right)"""
        return other ** self._getval()

    def __rsub__(self, other):
        """- (right)"""
        return other - self._getval()


def _log_value(value: float):
    if not np.isfinite(value):
        return value
    if value == 1:
        value += 1e-10
    return np.log(value)


# A reexp for ONLY matching scientific
_match_scientific = re.compile(r"[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)")


def _sanatize_parameter_list(li: list) -> list:
    for i, value in enumerate(li):
        if isinstance(value, str) and _match_scientific.match(value):
            li[i] = float(value)

    return li


def _retrieve_from_list_by_type(li: list, t: type | tuple[type, ...], default: Any):
    tmp = list(filter(lambda x: isinstance(x, t), li))
    if not tmp:
        return default
    li.remove(tmp[0])
    return tmp[0]



from __future__ import annotations

import xarray as xr

from glotaran.analysis.problem import Problem
from glotaran.builtin.models.kinetic_image.irf import IrfMultiGaussian
from glotaran.builtin.models.kinetic_image.kinetic_baseline_megacomplex import (
    KineticBaselineMegacomplex,
)
from glotaran.builtin.models.kinetic_image.kinetic_image_result import (
    retrieve_decay_associated_data,
)
from glotaran.builtin.models.kinetic_image.kinetic_image_result import retrieve_irf
from glotaran.builtin.models.kinetic_image.kinetic_image_result import (
    retrieve_species_associated_data,
)
from glotaran.builtin.models.kinetic_spectrum.coherent_artifact_megacomplex import (
    CoherentArtifactMegacomplex,
)
from glotaran.builtin.models.kinetic_spectrum.spectral_irf import IrfSpectralMultiGaussian


def finalize_kinetic_spectrum_result(model, problem: Problem, data: dict[str, xr.Dataset]):

    for label, dataset in data.items():

        dataset_descriptor = problem.filled_dataset_descriptors[label]

        if any(
            isinstance(megacomplex, KineticBaselineMegacomplex)
            for megacomplex in dataset_descriptor.megacomplex
        ):
            dataset["baseline"] = dataset.clp.sel(clp_label=f"{dataset_descriptor.label}_baseline")

        retrieve_species_associated_data(problem.model, dataset, dataset_descriptor, "spectra")

        retrieve_decay_associated_data(problem.model, dataset, dataset_descriptor, "spectra")

        irf = dataset_descriptor.irf
        if isinstance(irf, IrfMultiGaussian):
            if isinstance(irf.center, list):
                dataset["irf_center"] = irf.center[0].value
                dataset["irf_width"] = irf.width[0].value
            else:
                dataset["irf_center"] = irf.center.value
                dataset["irf_width"] = irf.width.value
        elif isinstance(irf, IrfSpectralMultiGaussian):

            dataset["irf"] = (
                ("time"),
                irf.calculate(0, dataset.coords["spectral"], dataset.coords["time"]),
            )

            if irf.dispersion_center:
                for i, dispersion in enumerate(
                    irf.calculate_dispersion(dataset.coords["spectral"].values)
                ):
                    dataset[f"center_dispersion_{i+1}"] = (
                        problem.model.global_dimension,
                        dispersion,
                    )
        else:
            retrieve_irf(problem.model, dataset, dataset_descriptor, "images")

        if any(
            isinstance(megacomplex, CoherentArtifactMegacomplex)
            for megacomplex in dataset_descriptor.megacomplex
        ):
            coherent_artifact = [
                c
                for c in dataset_descriptor.megacomplex
                if isinstance(c, CoherentArtifactMegacomplex)
            ][0]
            dataset.coords["coherent_artifact_order"] = list(range(1, coherent_artifact.order + 1))
            dataset["coherent_artifact_concentration"] = (
                (problem.model.model_dimension, "coherent_artifact_order"),
                dataset.matrix.sel(clp_label=coherent_artifact.compartments()).values,
            )
            dataset["coherent_artifact_associated_spectra"] = (
                (problem.model.global_dimension, "coherent_artifact_order"),
                dataset.clp.sel(clp_label=coherent_artifact.compartments()).values,
            )



from __future__ import annotations

from typing import Any
from typing import Mapping

import numpy as np

from glotaran.builtin.models.kinetic_image.initial_concentration import InitialConcentration
from glotaran.builtin.models.kinetic_image.irf import Irf
from glotaran.builtin.models.kinetic_image.k_matrix import KMatrix
from glotaran.builtin.models.kinetic_image.kinetic_image_dataset_descriptor import (
    KineticImageDatasetDescriptor,
)
from glotaran.builtin.models.kinetic_image.kinetic_image_matrix import kinetic_image_matrix
from glotaran.builtin.models.kinetic_image.kinetic_image_megacomplex import KineticImageMegacomplex
from glotaran.builtin.models.kinetic_image.kinetic_image_result import (
    finalize_kinetic_image_result,
)
from glotaran.model import Model
from glotaran.model import model

class KineticImageModel(Model):
    dataset: Mapping[str, KineticImageDatasetDescriptor]
    megacomplex: Mapping[str, KineticImageMegacomplex]
    @staticmethod
    def matrix(  # type: ignore[override]
        dataset_descriptor: KineticImageDatasetDescriptor = ..., axis=..., index=..., irf=...
    ) -> tuple[None, None] | tuple[list[Any], np.ndarray]: ...
    @property
    def initial_concentration(self) -> Mapping[str, InitialConcentration]: ...
    @property
    def k_matrix(self) -> Mapping[str, KMatrix]: ...
    @property
    def irf(self) -> Mapping[str, Irf]: ...



from __future__ import annotations

import xarray as xr

from glotaran.io import DataIoInterface
from glotaran.io import register_data_io
from glotaran.project import SavingOptions
from glotaran.project import default_data_filters


@register_data_io("nc")
class NetCDFDataIo(DataIoInterface):
    def load_dataset(self, file_name: str) -> xr.Dataset | xr.DataArray:
        return xr.open_dataset(file_name)

    def save_dataset(
        self,
        dataset: xr.Dataset,
        file_name: str,
        *,
        saving_options: SavingOptions = SavingOptions(),
    ):

        data_to_save = dataset

        data_filter = (
            saving_options.data_filter
            if saving_options.data_filter is not None
            else default_data_filters[saving_options.level]
        )

        if data_filter is not None:

            data_to_save = xr.Dataset()
            for item in data_filter:
                data_to_save[item] = dataset[item]

        data_to_save.to_netcdf(file_name)



import numpy as np
import xarray as xr

from glotaran.analysis.optimize import optimize
from glotaran.analysis.simulation import simulate
from glotaran.analysis.util import calculate_matrix
from glotaran.builtin.models.kinetic_spectrum import KineticSpectrumModel
from glotaran.parameter import ParameterGroup
from glotaran.project import Scheme


def test_coherent_artifact():
    model_dict = {
        "initial_concentration": {
            "j1": {"compartments": ["s1"], "parameters": ["2"]},
        },
        "megacomplex": {
            "mc1": {"type": "kinetic-decay", "k_matrix": ["k1"]},
            "mc2": {"type": "coherent-artifact", "order": 3},
        },
        "k_matrix": {
            "k1": {
                "matrix": {
                    ("s1", "s1"): "1",
                }
            }
        },
        "irf": {
            "irf1": {
                "type": "spectral-gaussian",
                "center": "2",
                "width": "3",
            },
        },
        "dataset": {
            "dataset1": {
                "initial_concentration": "j1",
                "megacomplex": ["mc1", "mc2"],
                "irf": "irf1",
            },
        },
    }
    model = KineticSpectrumModel.from_dict(model_dict.copy())

    parameters = ParameterGroup.from_list(
        [
            101e-4,
            [10, {"vary": False, "non-negative": False}],
            [20, {"vary": False, "non-negative": False}],
            [30, {"vary": False, "non-negative": False}],
        ]
    )

    time = np.asarray(np.arange(0, 50, 1.5))

    data = model.dataset["dataset1"].fill(model, parameters)
    compartments, matrix = calculate_matrix(model, data, {}, {"time": time})

    assert len(compartments) == 4
    for i in range(1, 4):
        assert compartments[i] == f"coherent_artifact_{i}"

    assert matrix.shape == (time.size, 4)

    clp = xr.DataArray(
        [[1, 1, 1, 1]],
        coords=[
            ("spectral", [0]),
            (
                "clp_label",
                [
                    "s1",
                    "coherent_artifact_1",
                    "coherent_artifact_2",
                    "coherent_artifact_3",
                ],
            ),
        ],
    )
    axis = {"time": time, "spectral": clp.spectral.data}
    data = simulate(model, "dataset1", parameters, axis, clp)

    dataset = {"dataset1": data}
    scheme = Scheme(
        model=model, parameters=parameters, data=dataset, maximum_number_function_evaluations=20
    )
    result = optimize(scheme)
    print(result.optimized_parameters)

    for label, param in result.optimized_parameters.all():
        assert np.allclose(param.value, parameters.get(label).value, rtol=1e-1)

    resultdata = result.data["dataset1"]
    assert np.array_equal(data.time, resultdata.time)
    assert np.array_equal(data.spectral, resultdata.spectral)
    assert data.data.shape == resultdata.data.shape
    assert data.data.shape == resultdata.fitted_data.shape
    assert np.allclose(data.data, resultdata.fitted_data, rtol=1e-2)

    assert "coherent_artifact_concentration" in resultdata
    assert resultdata["coherent_artifact_concentration"].shape == (time.size, 3)

    assert "coherent_artifact_associated_spectra" in resultdata
    assert resultdata["coherent_artifact_associated_spectra"].shape == (1, 3)



"""Test deprecated functionality in 'glotaran.model.base_model'."""
from __future__ import annotations

from glotaran.deprecation.modules.test import deprecation_warning_on_call_test_helper
from glotaran.model.base_model import Model


def test_model_simulate_method():
    """Model.simulate raises deperecation warning"""
    deprecation_warning_on_call_test_helper(Model().simulate)



""" Glotaran Spectral Relation """
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING
from typing import List
from typing import Tuple

import numpy as np
import xarray as xr

from glotaran.model import model_attribute
from glotaran.parameter import Parameter

if TYPE_CHECKING:
    from typing import Any

    from glotaran.builtin.models.kinetic_spectrum.kinetic_spectrum_model import (
        KineticSpectrumModel,
    )
    from glotaran.parameter import ParameterGroup


@model_attribute(
    properties={
        "compartment": str,
        "target": str,
        "parameter": Parameter,
        "interval": List[Tuple[float, float]],
    },
    no_label=True,
)
class SpectralRelation:
    def applies(self, index: Any) -> bool:
        """
        Returns true if the index is in one of the intervals.

        Parameters
        ----------
        index :

        Returns
        -------
        applies : bool

        """
        return any(interval[0] <= index <= interval[1] for interval in self.interval)


def create_spectral_relation_matrix(
    model: KineticSpectrumModel,
    dataset: str,
    parameters: ParameterGroup,
    clp_labels: list[str],
    matrix: np.ndarray,
    index: float,
) -> tuple[list[str], np.ndarray]:
    relation_matrix = np.diagflat([1.0 for _ in clp_labels])

    idx_to_delete = []
    for relation in model.spectral_relations:
        if relation.target in clp_labels and relation.applies(index):

            if relation.compartment not in clp_labels:
                warnings.warn(
                    "Relation between compartments '{relation.compartment}' and "
                    f"'{relation.target}' cannot be applied for '{dataset}'. "
                    f" '{relation.source}' not in dataset"
                )
                continue

            relation = relation.fill(model, parameters)
            source_idx = clp_labels.index(relation.compartment)
            target_idx = clp_labels.index(relation.target)
            relation_matrix[target_idx, source_idx] = relation.parameter
            idx_to_delete.append(target_idx)

    clp_labels = [label for i, label in enumerate(clp_labels) if i not in idx_to_delete]
    relation_matrix = np.delete(relation_matrix, idx_to_delete, axis=1)
    return (clp_labels, relation_matrix)


def apply_spectral_relations(
    model: KineticSpectrumModel,
    dataset: str,
    parameters: ParameterGroup,
    clp_labels: list[str],
    matrix: np.ndarray,
    index: float,
) -> tuple[list[str], np.ndarray]:

    if not model.spectral_relations:
        return (clp_labels, matrix)

    reduced_clp_labels, relation_matrix = create_spectral_relation_matrix(
        model, dataset, parameters, clp_labels, matrix, index
    )

    reduced_matrix = matrix @ relation_matrix

    return (reduced_clp_labels, reduced_matrix)


def retrieve_related_clps(
    model: KineticSpectrumModel,
    parameters: ParameterGroup,
    clp_labels: dict[str, list[str] | list[list[str]]],
    clps: dict[str, list[np.ndarray]],
    data: dict[str, xr.Dataset],
) -> dict[str, list[np.ndarray]]:

    for relation in model.spectral_relations:
        relation = relation.fill(model, parameters)
        for label, dataset_clp_labels in clp_labels.items():
            for i, index in enumerate(data[label].coords[model.global_dimension]):
                if (
                    relation.target in dataset_clp_labels[i]
                    and relation.compartment in dataset_clp_labels[i]
                    and relation.applies(index)
                ):
                    target_idx = dataset_clp_labels[i].index(relation.target)
                    source_idx = dataset_clp_labels[i].index(relation.compartment)
                    clps[label][i][target_idx] = clps[label][i][source_idx] * relation.parameter

    return clps



"""The parameter group class"""

from __future__ import annotations

from copy import copy
from textwrap import indent
from typing import Generator

import asteval
import numpy as np
import pandas as pd
from tabulate import tabulate

from glotaran.parameter.parameter import Parameter
from glotaran.utils.ipython import MarkdownStr


class ParameterNotFoundException(Exception):
    """Raised when a Parameter is not found in the Group."""

    def __init__(self, path, label):
        super().__init__(f"Cannot find parameter {'.'.join(path+[label])!r}")


class ParameterGroup(dict):
    def __init__(self, label: str = None, root_group: ParameterGroup = None):
        """Represents are group of parameters. Can contain other groups, creating a
        tree-like hierarchy.

        Parameters
        ----------
        label :
            The label of the group.
        """
        if label is not None and not Parameter.valid_label(label):
            raise ValueError(f"'{label}' is not a valid group label.")
        self._label = label
        self._parameters = {}
        self._root_group = root_group
        self._evaluator = (
            asteval.Interpreter(symtable=asteval.make_symbol_table(group=self))
            if root_group is None
            else None
        )
        super().__init__()

    @classmethod
    def from_dict(
        cls,
        parameter_dict: dict[str, dict | list],
        label: str = None,
        root_group: ParameterGroup = None,
    ) -> ParameterGroup:
        """Creates a :class:`ParameterGroup` from a dictionary.

        Parameters
        ----------
        parameter_dict :
            A parameter dictionary containing parameters.
        label :
            The label of root group.
        root_group:
            The root group
        """
        root = cls(label=label, root_group=root_group)
        for label, item in parameter_dict.items():
            label = str(label)
            if isinstance(item, dict):
                root.add_group(cls.from_dict(item, label=label, root_group=root))
            if isinstance(item, list):
                root.add_group(cls.from_list(item, label=label, root_group=root))
        if root_group is None:
            root.update_parameter_expression()
        return root

    @classmethod
    def from_list(
        cls,
        parameter_list: list[float | list],
        label: str = None,
        root_group: ParameterGroup = None,
    ) -> ParameterGroup:
        """Creates a :class:`ParameterGroup` from a list.

        Parameters
        ----------
        parameter_list :
            A parameter list containing parameters
        label :
            The label of the root group.
        root_group:
            The root group
        """
        root = cls(label=label, root_group=root_group)

        # get defaults
        defaults = None
        for item in parameter_list:
            if isinstance(item, dict):
                defaults = item
                break

        for i, item in enumerate(parameter_list):
            if isinstance(item, (str, int, float)):
                try:
                    item = float(item)
                except Exception:
                    pass
            if isinstance(item, (float, int, list)):
                root.add_parameter(
                    Parameter.from_list_or_value(item, label=str(i + 1), default_options=defaults)
                )
        if root_group is None:
            root.update_parameter_expression()
        return root

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame, source: str = "DataFrame") -> ParameterGroup:
        """Creates a :class:`ParameterGroup` from a :class:`pandas.DataFrame`"""

        for column_name in ["label", "value"]:
            if column_name not in df:
                raise ValueError(f"Missing column '{column_name}' in '{source}'")

        for column_name in ["minimum", "maximum", "value"]:
            if column_name in df and any(not np.isreal(v) for v in df[column_name]):
                raise ValueError(f"Column '{column_name}' in '{source}' has non numeric values")

        for column_name in ["non-negative", "vary"]:
            if column_name in df and any(not isinstance(v, bool) for v in df[column_name]):
                raise ValueError(f"Column '{column_name}' in '{source}' has non boolean values")

        root = cls()

        for i, full_label in enumerate(df["label"]):
            path = full_label.split(".")
            group = root
            while len(path) > 1:
                group_label = path.pop(0)
                if group_label not in group:
                    group.add_group(ParameterGroup(label=group_label, root_group=group))
                group = group[group_label]
            label = path.pop()
            value = df["value"][i]
            minimum = df["minimum"][i] if "minimum" in df else -np.inf
            maximum = df["maximum"][i] if "maximum" in df else np.inf
            non_negative = df["non-negative"][i] if "non-negative" in df else False
            vary = df["vary"][i] if "vary" in df else True
            expression = (
                df["expression"][i]
                if "expression" in df and isinstance(df["expression"][i], str)
                else None
            )

            parameter = Parameter(
                label=label,
                full_label=full_label,
                value=value,
                expression=expression,
                maximum=maximum,
                minimum=minimum,
                non_negative=non_negative,
                vary=vary,
            )
            group.add_parameter(parameter)
        root.update_parameter_expression()
        return root

    @property
    def label(self) -> str:
        """Label of the group."""
        return self._label

    @property
    def root_group(self) -> ParameterGroup:
        """Root of the group."""
        return self._root_group

    def to_dataframe(self) -> pd.DataFrame:
        parameter_dict = {
            "label": [],
            "value": [],
            "minimum": [],
            "maximum": [],
            "vary": [],
            "non-negative": [],
            "expression": [],
        }
        for label, parameter in self.all():
            parameter_dict["label"].append(label)
            parameter_dict["value"].append(parameter.value)
            parameter_dict["minimum"].append(parameter.minimum)
            parameter_dict["maximum"].append(parameter.maximum)
            parameter_dict["vary"].append(parameter.vary)
            parameter_dict["non-negative"].append(parameter.non_negative)
            parameter_dict["expression"].append(parameter.expression)
        return pd.DataFrame(parameter_dict)

    def to_csv(self, filename: str, delimiter: str = ","):
        """Writes a :class:`ParameterGroup` to a CSV file.

        Parameters
        ----------
        filepath :
            The path to the CSV file.
        delimiter : str
            The delimiter of the CSV file.
        """
        self.to_dataframe().to_csv(filename, sep=delimiter, na_rep="None", index=False)

    def add_parameter(self, parameter: Parameter | list[Parameter]):
        """Adds a :class:`Parameter` to the group.

        Parameters
        ----------
        parameter :
            The parameter to add.
        """
        if not isinstance(parameter, list):
            parameter = [parameter]
        if any(not isinstance(p, Parameter) for p in parameter):
            raise TypeError("Parameter must be  instance of glotaran.parameter.Parameter")
        for p in parameter:
            p.index = len(self._parameters) + 1
            if p.label is None:
                p.label = f"{p.index}"
            p.full_label = f"{self.label}.{p.label}" if self.label else p.label
            self._parameters[p.label] = p

    def add_group(self, group: ParameterGroup):
        """Adds a :class:`ParameterGroup` to the group.

        Parameters
        ----------
        group :
            The group to add.
        """
        if not isinstance(group, ParameterGroup):
            raise TypeError("Group must be glotaran.parameter.ParameterGroup")
        self[group.label] = group

    def get_nr_roots(self) -> int:
        """Returns the number of roots of the group."""
        n = 0
        root = self.root_group
        while root is not None:
            n += 1
            root = root.root_group
        return n

    def groups(self) -> Generator[ParameterGroup, None, None]:
        """Returns a generator over all groups and their subgroups."""
        for group in self:
            yield from group.groups()

    def has(self, label: str) -> bool:
        """Checks if a parameter with the given label is in the group or in a subgroup.

        Parameters
        ----------
        label :
            The label of the parameter, with its path in a parameter group prepended.
        """

        try:
            self.get(label)
            return True
        except Exception:
            return False

    def get(self, label: str) -> Parameter:
        """Gets a :class:`Parameter` by its label.

        Parameters
        ----------
        label :
            The label of the parameter, with its path in a parameter group prepended.
        """

        # sometimes the spec parser delivers the labels as int
        label = str(label)

        path = label.split(".")
        label = path.pop()

        # TODO: audit this code
        group = self
        for element in path:
            try:
                group = group[element]
            except KeyError:
                raise ParameterNotFoundException(path, label)
        try:
            return group._parameters[label]
        except KeyError:
            raise ParameterNotFoundException(path, label)

    def copy(self) -> ParameterGroup:
        root = ParameterGroup(label=self.label, root_group=self.root_group)

        for label, parameter in self._parameters.items():
            root._parameters[label] = copy(parameter)

        for label, group in self.items():
            root[label] = group.copy()

        return root

    def all(
        self, root: str = None, separator: str = "."
    ) -> Generator[tuple[str, Parameter], None, None]:
        """Returns a generator over all parameter in the group and it's subgroups together with
        their labels.

        Parameters
        ----------
        root :
            The label of the root group
        separator:
            The separator for the parameter labels.
        """

        root = f"{root}{self.label}{separator}" if root is not None else ""
        for label, p in self._parameters.items():
            yield (f"{root}{label}", p)
        for _, l in self.items():
            yield from l.all(root=root, separator=separator)

    def get_label_value_and_bounds_arrays(
        self, exclude_non_vary: bool = False
    ) -> tuple[list[str], np.ndarray, np.ndarray, np.ndarray]:
        """Returns a arrays of all parameter labels, values and bounds.

        Parameters
        ----------

        exclude_non_vary: bool = False
            If true, parameters with `vary=False` are excluded.
        """
        self.update_parameter_expression()

        labels = []
        values = []
        lower_bounds = []
        upper_bounds = []

        for label, parameter in self.all():
            if not exclude_non_vary or parameter.vary:
                labels.append(label)
                value, minimum, maximum = parameter.get_value_and_bounds_for_optimization()
                values.append(value)
                lower_bounds.append(minimum)
                upper_bounds.append(maximum)

        return labels, np.asarray(values), np.asarray(lower_bounds), np.asarray(upper_bounds)

    def set_from_label_and_value_arrays(self, labels: list[str], values: np.ndarray):
        """Updates the parameter values from a list of labels and values."""

        if len(labels) != len(values):
            raise ValueError(
                f"Length of labels({len(labels)}) not equal to length of values({len(values)})."
            )

        for label, value in zip(labels, values):
            self.get(label).set_value_from_optimization(value)

        self.update_parameter_expression()

    def update_parameter_expression(self):
        """Updates all parameters which have an expression."""
        for label, parameter in self.all():
            if parameter.expression is not None:
                value = self._evaluator(parameter.transformed_expression)
                if not isinstance(value, (int, float)):
                    raise ValueError(
                        f"Expression '{parameter.expression}' of parameter '{label}' evaluates to"
                        f"non numeric value '{value}'."
                    )
                parameter.value = value

    def markdown(self) -> MarkdownStr:
        """Formats the :class:`ParameterGroup` as markdown string.

        This is done by recursing the nested :class:`ParameterGroup` tree.
        """
        node_indentation = "  " * self.get_nr_roots()
        return_string = ""
        table_header = [
            "_Label_",
            "_Value_",
            "_StdErr_",
            "_Min_",
            "_Max_",
            "_Vary_",
            "_Non-Negative_",
            "_Expr_",
        ]
        if self.label is not None:
            return_string += f"{node_indentation}* __{self.label}__:\n"
        if len(self._parameters):
            parameter_rows = [
                [
                    parameter.label,
                    parameter.value,
                    parameter.standard_error,
                    parameter.minimum,
                    parameter.maximum,
                    parameter.vary,
                    parameter.non_negative,
                    parameter.expression,
                ]
                for _, parameter in self._parameters.items()
            ]
            parameter_table = indent(
                tabulate(
                    parameter_rows, headers=table_header, tablefmt="github", missingval="None"
                ),
                f"  {node_indentation}",
            )
            return_string += f"\n{parameter_table}\n\n"
        for _, child_group in sorted(self.items()):
            return_string += f"{child_group.__str__()}"
        return MarkdownStr(return_string)

    def _repr_markdown_(self) -> str:
        """Special method used by ``ipython`` to render markdown."""
        return str(self.markdown())

    def __repr__(self):
        """Representation used by repl and tracebacks."""
        if self.label is None:
            return f"{type(self).__name__}.from_dict({super().__repr__()})"
        elif len(self._parameters):
            parameter_short_notations = [
                f"[{parameter.label!r}, {parameter.value}]"
                for _, parameter in self._parameters.items()
            ]
            return f"[{', '.join(parameter_short_notations)}]"
        else:
            return super().__repr__()

    def __str__(self):
        """Representation used by print and str."""
        return str(self.markdown())



"""Implementation of the folder Io plugin.

The current implementation is an exact copy of how ``Result.save(path)``
worked in glotaran 0.3.x and meant as an compatibility function.
"""

from __future__ import annotations

import os
from typing import TYPE_CHECKING

from glotaran.io.interface import ProjectIoInterface
from glotaran.plugin_system.project_io_registration import register_project_io

if TYPE_CHECKING:
    from glotaran.project import Result


@register_project_io(["folder", "legacy"])
class FolderProjectIo(ProjectIoInterface):
    """Project Io plugin to save result data to a folder.

    There won't be a serialization of the Result object, but simply
    a markdown summary output and the important data saved to files.
    """

    def save_result(self, result: Result, result_path: str) -> list[str]:
        """Save the result to a given folder.

        Returns a list with paths of all saved items.
        The following files are saved:
        * `result.md`: The result with the model formatted as markdown text.
        * `optimized_parameters.csv`: The optimized parameter as csv file.
        * `{dataset_label}.nc`: The result data for each dataset as NetCDF file.

        Parameters
        ----------
        result : Result
            Result instance to be saved.
        result_path : str
            The path to the folder in which to save the result.

        Returns
        -------
        list[str]
            List of file paths which were created.

        Raises
        ------
        ValueError
            If ``result_path`` is a file.
        """
        if not os.path.exists(result_path):
            os.makedirs(result_path)
        if not os.path.isdir(result_path):
            raise ValueError(f"The path '{result_path}' is not a directory.")

        paths = []

        md_path = os.path.join(result_path, "result.md")
        with open(md_path, "w") as f:
            f.write(str(result.markdown()))
        paths.append(md_path)

        csv_path = os.path.join(result_path, "optimized_parameters.csv")
        result.optimized_parameters.to_csv(csv_path)
        paths.append(csv_path)

        for label, data in result.data.items():
            nc_path = os.path.join(result_path, f"{label}.nc")
            data.to_netcdf(nc_path, engine="netcdf4")
            paths.append(nc_path)

        return paths



import numpy as np
import xarray as xr

from glotaran.analysis.optimize import optimize
from glotaran.analysis.simulation import simulate
from glotaran.analysis.util import calculate_matrix
from glotaran.builtin.models.kinetic_spectrum import KineticSpectrumModel
from glotaran.builtin.models.kinetic_spectrum.spectral_relations import (
    create_spectral_relation_matrix,
)
from glotaran.parameter import ParameterGroup
from glotaran.project import Scheme


def test_spectral_relation():
    model = KineticSpectrumModel.from_dict(
        {
            "initial_concentration": {
                "j1": {
                    "compartments": ["s1", "s2", "s3", "s4"],
                    "parameters": ["i.1", "i.2", "i.3", "i.4"],
                },
            },
            "megacomplex": {
                "mc1": {"k_matrix": ["k1"]},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s1", "s1"): "kinetic.1",
                        ("s2", "s2"): "kinetic.1",
                        ("s3", "s3"): "kinetic.1",
                        ("s4", "s4"): "kinetic.1",
                    }
                }
            },
            "spectral_relations": [
                {
                    "compartment": "s1",
                    "target": "s2",
                    "parameter": "rel.1",
                    "interval": [(0, 2)],
                },
                {
                    "compartment": "s1",
                    "target": "s3",
                    "parameter": "rel.2",
                    "interval": [(0, 2)],
                },
            ],
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "megacomplex": ["mc1"],
                },
            },
        }
    )
    print(model)

    rel1, rel2 = 10, 20
    parameters = ParameterGroup.from_dict(
        {
            "kinetic": [1e-4],
            "i": [1, 2, 3, 4],
            "rel": [rel1, rel2],
        }
    )

    time = np.asarray(np.arange(0, 50, 1.5))
    dataset = model.dataset["dataset1"].fill(model, parameters)
    compartments, matrix = calculate_matrix(model, dataset, {}, {"time": time})

    assert len(compartments) == 4
    assert matrix.shape == (time.size, 4)

    reduced_compartments, relation_matrix = create_spectral_relation_matrix(
        model, "dataset1", parameters, compartments, matrix, 1
    )

    print(relation_matrix)
    assert len(reduced_compartments) == 2
    assert relation_matrix.shape == (4, 2)
    assert np.array_equal(
        relation_matrix,
        [
            [1.0, 0.0],
            [10.0, 0.0],
            [20.0, 0.0],
            [0.0, 1.0],
        ],
    )

    reduced_compartments, reduced_matrix = model.constrain_matrix_function(
        "dataset1", parameters, compartments, matrix, 1
    )

    assert reduced_matrix.shape == (time.size, 2)

    print(reduced_matrix[0, 0], matrix[0, 0], matrix[0, 1], matrix[0, 2])
    assert np.allclose(
        reduced_matrix[:, 0], matrix[:, 0] + rel1 * matrix[:, 1] + rel2 * matrix[:, 2]
    )

    clp = xr.DataArray(
        [[1.0, 10.0, 20.0, 1]], coords=(("spectral", [1]), ("clp_label", ["s1", "s2", "s3", "s4"]))
    )

    data = simulate(
        model, "dataset1", parameters, clp=clp, axes={"time": time, "spectral": np.array([1])}
    )

    dataset = {"dataset1": data}
    scheme = Scheme(
        model=model, parameters=parameters, data=dataset, maximum_number_function_evaluations=20
    )
    result = optimize(scheme)

    for label, param in result.optimized_parameters.all():
        if param.vary:
            assert np.allclose(param.value, parameters.get(label).value, rtol=1e-1)

    result_data = result.data["dataset1"]
    print(result_data.species_associated_spectra)
    assert result_data.species_associated_spectra.shape == (1, 4)
    assert (
        result_data.species_associated_spectra[0, 1]
        == rel1 * result_data.species_associated_spectra[0, 0]
    )
    assert np.allclose(
        result_data.species_associated_spectra[0, 2].values,
        rel2 * result_data.species_associated_spectra[0, 0].values,
    )


if __name__ == "__main__":
    test_spectral_relation()



"""Test deprecated functionality in 'glotaran.project.result'."""
from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from glotaran.deprecation.modules.test import deprecation_warning_on_call_test_helper
from glotaran.project.test.test_result import dummy_result  # noqa: F401

if TYPE_CHECKING:
    from py.path import local as LocalPath

    from glotaran.project.result import Result


def test_Result_save_method(tmpdir: LocalPath, dummy_result: Result):  # noqa: F811
    """Result.save(result_dir) creates all file"""
    result_dir = tmpdir / "dummy"
    result_dir.mkdir()

    deprecation_warning_on_call_test_helper(
        dummy_result.save, args=[str(result_dir)], raise_exception=True
    )

    assert (result_dir / "result.md").exists()
    assert (result_dir / "optimized_parameters.csv").exists()
    assert (result_dir / "dataset1.nc").exists()
    assert (result_dir / "dataset2.nc").exists()
    assert (result_dir / "dataset3.nc").exists()


def test_Result_get_dataset_method(dummy_result: Result):  # noqa: F811
    """Result.get_dataset(dataset_label) gives correct dataset."""

    result = deprecation_warning_on_call_test_helper(
        dummy_result.get_dataset, args=["dataset1"], raise_exception=True
    )

    assert result == dummy_result.data["dataset1"]


def test_Result_get_dataset_method_error(dummy_result: Result):  # noqa: F811
    """Result.get_dataset(dataset_label) error on wrong key."""

    with pytest.raises(ValueError, match="Unknown dataset 'foo'"):
        deprecation_warning_on_call_test_helper(
            dummy_result.get_dataset, args=["foo"], raise_exception=True
        )



import collections

import numpy as np
import pytest
import xarray as xr

from glotaran.analysis.problem import Problem
from glotaran.analysis.problem_grouped import GroupedProblem
from glotaran.analysis.problem_ungrouped import UngroupedProblem
from glotaran.analysis.simulation import simulate
from glotaran.analysis.test.models import MultichannelMulticomponentDecay as suite
from glotaran.analysis.test.models import SimpleTestModel
from glotaran.parameter import ParameterGroup
from glotaran.project import Scheme


@pytest.fixture(
    scope="module", params=[[True, True], [True, False], [False, True], [False, False]]
)
def problem(request) -> Problem:
    model = suite.model
    model.is_grouped = request.param[0]
    model.is_index_dependent = request.param[1]

    dataset = simulate(
        suite.sim_model,
        "dataset1",
        suite.wanted_parameters,
        {"e": suite.e_axis, "c": suite.c_axis},
    )
    scheme = Scheme(model=model, parameters=suite.initial_parameters, data={"dataset1": dataset})
    return GroupedProblem(scheme) if request.param[0] else UngroupedProblem(scheme)


def test_problem(problem: Problem):
    assert problem.grouped == problem.model.is_grouped
    assert problem.index_dependent == problem.model.is_index_dependent


def test_problem_bag(problem: Problem):

    bag = problem.bag

    if problem.grouped:
        assert isinstance(bag, collections.deque)
        assert len(bag) == suite.e_axis.size
        assert problem.groups == {"dataset1": ["dataset1"]}
    else:
        assert isinstance(bag, dict)
        assert "dataset1" in bag


def test_problem_matrices(problem: Problem):
    problem.calculate_matrices()

    if problem.grouped:
        if problem.index_dependent:
            assert all(isinstance(m, list) for m in problem.reduced_clp_labels)
            assert all(isinstance(m, np.ndarray) for m in problem.reduced_matrices)
            assert len(problem.reduced_clp_labels) == suite.e_axis.size
            assert len(problem.reduced_matrices) == suite.e_axis.size
        else:
            assert "dataset1" in problem.reduced_clp_labels
            assert "dataset1" in problem.reduced_matrices
            assert isinstance(problem.reduced_clp_labels["dataset1"], list)
            assert isinstance(problem.reduced_matrices["dataset1"], np.ndarray)
    else:
        if problem.index_dependent:
            assert isinstance(problem.reduced_clp_labels, dict)
            assert isinstance(problem.reduced_matrices, dict)
            assert isinstance(problem.reduced_matrices["dataset1"], list)
            assert all(isinstance(c, list) for c in problem.reduced_clp_labels["dataset1"])
            assert all(isinstance(m, np.ndarray) for m in problem.reduced_matrices["dataset1"])
        else:
            assert isinstance(problem.reduced_matrices["dataset1"], np.ndarray)

        assert isinstance(problem.clp_labels, dict)
        assert isinstance(problem.matrices, dict)
        assert isinstance(problem.reduced_clp_labels["dataset1"], list)
        assert "dataset1" in problem.reduced_clp_labels
        assert "dataset1" in problem.reduced_matrices


def test_problem_residuals(problem: Problem):
    problem.calculate_residual()
    if problem.grouped:
        assert isinstance(problem.residuals, list)
        assert all(isinstance(r, np.ndarray) for r in problem.residuals)
        assert len(problem.residuals) == suite.e_axis.size
    else:
        assert isinstance(problem.residuals, dict)
        assert "dataset1" in problem.residuals
        assert all(isinstance(r, np.ndarray) for r in problem.residuals["dataset1"])
        assert len(problem.residuals["dataset1"]) == suite.e_axis.size
    assert isinstance(problem.reduced_clps, dict)
    assert "dataset1" in problem.reduced_clps
    assert all(isinstance(c, np.ndarray) for c in problem.reduced_clps["dataset1"])
    assert len(problem.reduced_clps["dataset1"]) == suite.e_axis.size
    assert isinstance(problem.clps, dict)
    assert "dataset1" in problem.clps
    assert all(isinstance(c, np.ndarray) for c in problem.clps["dataset1"])
    assert len(problem.clps["dataset1"]) == suite.e_axis.size
    assert isinstance(problem.additional_penalty, np.ndarray)
    assert problem.additional_penalty.size == 1
    assert problem.additional_penalty[0] == 0.1
    assert isinstance(problem.full_penalty, np.ndarray)
    assert (
        problem.full_penalty.size
        == (suite.c_axis.size * suite.e_axis.size) + problem.additional_penalty.size
    )


def test_problem_result_data(problem: Problem):

    data = problem.create_result_data()

    assert "dataset1" in data

    dataset = data["dataset1"]

    assert "clp_label" in dataset.coords
    assert np.array_equal(dataset.clp_label, ["s1", "s2", "s3", "s4"])

    assert problem.model.global_dimension in dataset.coords
    assert np.array_equal(dataset.coords[problem.model.global_dimension], suite.e_axis)

    assert problem.model.model_dimension in dataset.coords
    assert np.array_equal(dataset.coords[problem.model.model_dimension], suite.c_axis)

    assert "matrix" in dataset
    matrix = dataset.matrix
    if problem.index_dependent:
        assert len(matrix.shape) == 3
        assert matrix.shape[0] == suite.e_axis.size
        assert matrix.shape[1] == suite.c_axis.size
        assert matrix.shape[2] == 4
    else:
        assert len(matrix.shape) == 2
        assert matrix.shape[0] == suite.c_axis.size
        assert matrix.shape[1] == 4

    assert "clp" in dataset
    clp = dataset.clp
    assert len(clp.shape) == 2
    assert clp.shape[0] == suite.e_axis.size
    assert clp.shape[1] == 4

    assert "weighted_residual" in dataset
    assert dataset.data.shape == dataset.weighted_residual.shape

    assert "residual" in dataset
    assert dataset.data.shape == dataset.residual.shape

    assert "residual_singular_values" in dataset
    assert "weighted_residual_singular_values" in dataset


def test_prepare_data():
    model_dict = {
        "dataset": {
            "dataset1": {
                "megacomplex": [],
            },
        },
        "weights": [
            {
                "datasets": ["dataset1"],
                "global_interval": (np.inf, 200),
                "model_interval": (4, 8),
                "value": 0.5,
            },
        ],
    }
    model = SimpleTestModel.from_dict(model_dict)
    print(model.validate())
    assert model.valid()

    parameters = ParameterGroup.from_list([])

    global_axis = np.asarray(range(50, 300))
    model_axis = np.asarray(range(15))

    dataset = xr.DataArray(
        np.ones((global_axis.size, model_axis.size)),
        coords={"e": global_axis, "c": model_axis},
        dims=("e", "c"),
    )

    scheme = Scheme(model, parameters, {"dataset1": dataset})
    problem = Problem(scheme)

    data = problem.data["dataset1"]
    print(data)
    assert "data" in data
    assert "weight" in data

    assert data.data.shape == (model_axis.size, global_axis.size)
    assert data.data.shape == data.weight.shape
    assert np.all(data.weight.sel(e=slice(0, 200), c=slice(4, 8)).values == 0.5)
    assert np.all(data.weight.sel(c=slice(0, 3)).values == 1)

    model_dict["weights"].append(
        {
            "datasets": ["dataset1"],
            "value": 0.2,
        }
    )
    model = SimpleTestModel.from_dict(model_dict)
    print(model.validate())
    assert model.valid()

    scheme = Scheme(model, parameters, {"dataset1": dataset})
    problem = Problem(scheme)
    data = problem.data["dataset1"]
    assert np.all(data.weight.sel(e=slice(0, 200), c=slice(4, 8)).values == 0.5 * 0.2)
    assert np.all(data.weight.sel(c=slice(0, 3)).values == 0.2)

    with pytest.warns(
        UserWarning,
        match="Ignoring model weight for dataset 'dataset1'"
        " because weight is already supplied by dataset.",
    ):
        Problem(Scheme(model, parameters, {"dataset1": data}))



from __future__ import annotations

from typing import List

import numpy as np
import xarray as xr

from glotaran.model import DatasetDescriptor
from glotaran.model import Megacomplex
from glotaran.model import Model
from glotaran.model import model
from glotaran.model import model_attribute
from glotaran.parameter import Parameter
from glotaran.parameter import ParameterGroup


def calculate_e(dataset, axis):
    compartments = ["s1", "s2"]
    r_compartments = []
    array = np.zeros((axis.shape[0], len(compartments)))

    for i in range(len(compartments)):
        r_compartments.append(compartments[i])
        for j in range(axis.shape[0]):
            array[j, i] = (i + j) * axis[j]
    return (r_compartments, array)


@model_attribute()
class SimpleTestMegacomplex(Megacomplex):
    def calculate_matrix(self, model, dataset_descriptor, indices, axis, **kwargs):
        assert "c" in axis
        assert "e" in axis

        axis = axis["c"]
        compartments = ["s1", "s2"]
        r_compartments = []
        array = np.zeros((axis.shape[0], len(compartments)))

        for i in range(len(compartments)):
            r_compartments.append(compartments[i])
            for j in range(axis.shape[0]):
                array[j, i] = (i + j) * axis[j]
        return (r_compartments, array)


@model(
    "simple_test",
    attributes={},
    model_dimension="c",
    global_matrix=calculate_e,
    global_dimension="e",
    megacomplex_types=SimpleTestMegacomplex,
)
class SimpleTestModel(Model):
    pass


@model_attribute(properties={})
class SimpleKineticMegacomplex(Megacomplex):
    def calculate_matrix(self, model, dataset_descriptor, indices, axis, **kwargs):
        assert "c" in axis
        assert "e" in axis
        axis = axis["c"]
        kinpar = -1 * np.asarray(dataset_descriptor.kinetic)
        if dataset_descriptor.label == "dataset3":
            # this case is for the ThreeDatasetDecay test
            compartments = [f"s{i+2}" for i in range(len(kinpar))]
        else:
            compartments = [f"s{i+1}" for i in range(len(kinpar))]
        array = np.exp(np.outer(axis, kinpar))
        return (compartments, array)


def calculate_spectral_simple(dataset_descriptor, axis):
    kinpar = -1 * np.array(dataset_descriptor.kinetic)
    if dataset_descriptor.label == "dataset3":
        # this case is for the ThreeDatasetDecay test
        compartments = [f"s{i+2}" for i in range(len(kinpar))]
    else:
        compartments = [f"s{i+1}" for i in range(len(kinpar))]
    array = np.asarray([[1 for _ in range(axis.size)] for _ in compartments])
    return compartments, array.T


def calculate_spectral_gauss(dataset, axis):
    location = np.asarray(dataset.location)
    amp = np.asarray(dataset.amplitude)
    delta = np.asarray(dataset.delta)

    array = np.empty((location.size, axis.size), dtype=np.float64)

    for i in range(location.size):
        array[i, :] = amp[i] * np.exp(-np.log(2) * np.square(2 * (axis - location[i]) / delta[i]))
    compartments = [f"s{i+1}" for i in range(location.size)]
    return compartments, array.T


def constrain_matrix_function_typecheck(
    model: type[Model],
    label: str,
    parameters: ParameterGroup,
    clp_labels: list[str],
    matrix: np.ndarray,
    index: float,
):
    assert isinstance(label, str)
    assert isinstance(parameters, ParameterGroup)
    assert isinstance(clp_labels, list)
    assert all(isinstance(clp_label, str) for clp_label in clp_labels)
    assert isinstance(matrix, np.ndarray)

    if model.index_dependent():
        assert isinstance(index, float)
    else:
        assert index is None

    model.constrain_matrix_function_called = True

    return (clp_labels, matrix)


def retrieve_clp_typecheck(
    model: type[Model],
    parameters: ParameterGroup,
    clp_labels: dict[str, list[str] | list[list[str]]],
    reduced_clp_labels: dict[str, list[str] | list[list[str]]],
    reduced_clps: dict[str, list[np.ndarray]],
    data: dict[str, xr.Dataset],
) -> dict[str, list[np.ndarray]]:
    assert isinstance(parameters, ParameterGroup)

    assert isinstance(reduced_clps, dict)
    assert all(isinstance(dataset_clps, list) for dataset_clps in reduced_clps.values())

    assert all(
        [isinstance(index_clps, np.ndarray) for index_clps in dataset_clps]
        for dataset_clps in reduced_clps.values()
    )

    assert isinstance(data, dict)
    assert all(isinstance(label, str) for label in data)
    assert all(isinstance(dataset, xr.Dataset) for dataset in data.values())

    assert isinstance(clp_labels, dict)
    assert isinstance(reduced_clp_labels, dict)
    assert all(
        isinstance(dataset_clp_labels, list) for dataset_clp_labels in reduced_clp_labels.values()
    )
    assert all(
        [[isinstance(label, str) for label in index_labels] for index_labels in dataset_clp_labels]
        for dataset_clp_labels in reduced_clp_labels.values()
    )
    if model.index_dependent():
        for dataset_clp_labels in clp_labels.values():
            assert all(isinstance(index_label, list) for index_label in dataset_clp_labels)
            assert all(
                [isinstance(label, str) for label in index_label]
                for index_label in dataset_clp_labels
            )
        assert all(
            [isinstance(index_labels, list) for index_labels in dataset_clp_labels]
            for dataset_clp_labels in reduced_clp_labels.values()
        )

    else:
        for dataset_clp_labels in clp_labels.values():
            assert all(isinstance(label, str) for label in dataset_clp_labels)

    model.retrieve_clp_function_called = True

    return reduced_clps


def additional_penalty_typecheck(
    model: type[Model],
    parameters: ParameterGroup,
    clp_labels: dict[str, list[str] | list[list[str]]],
    clps: dict[str, list[np.ndarray]],
    matrices: dict[str, np.ndarray | list[np.ndarray]],
    data: dict[str, xr.Dataset],
    group_tolerance: float,
) -> np.ndarray:
    assert isinstance(parameters, ParameterGroup)
    assert isinstance(group_tolerance, float)

    assert isinstance(clps, dict)
    assert all(isinstance(dataset_clps, list) for dataset_clps in clps.values())
    assert all(
        [isinstance(index_clps, np.ndarray) for index_clps in dataset_clps]
        for dataset_clps in clps.values()
    )

    assert isinstance(data, dict)
    assert all(isinstance(label, str) for label in data)
    assert all(isinstance(dataset, xr.Dataset) for dataset in data.values())

    assert isinstance(clp_labels, dict)
    assert isinstance(matrices, dict)
    if model.index_dependent():
        for dataset_clp_labels in clp_labels.values():
            assert all(isinstance(index_label, list) for index_label in dataset_clp_labels)
            assert all(
                [isinstance(label, str) for label in index_label]
                for index_label in dataset_clp_labels
            )

        for matrix in matrices.values():
            assert isinstance(matrix, list)
            assert all(isinstance(index_matrix, np.ndarray) for index_matrix in matrix)
    else:
        for dataset_clp_labels in clp_labels.values():
            assert all(isinstance(label, str) for label in dataset_clp_labels)
        for matrix in matrices.values():
            assert isinstance(matrix, np.ndarray)

    model.additional_penalty_function_called = True

    return np.asarray([0.1])


@model_attribute(
    properties={
        "kinetic": List[Parameter],
    }
)
class DecayDatasetDescriptor(DatasetDescriptor):
    pass


@model_attribute(
    properties={
        "kinetic": List[Parameter],
        "location": {"type": List[Parameter], "allow_none": True},
        "amplitude": {"type": List[Parameter], "allow_none": True},
        "delta": {"type": List[Parameter], "allow_none": True},
    }
)
class GaussianShapeDecayDatasetDescriptor(DatasetDescriptor):
    pass


@model(
    "one_channel",
    attributes={},
    dataset_type=DecayDatasetDescriptor,
    model_dimension="c",
    global_matrix=calculate_spectral_simple,
    global_dimension="e",
    megacomplex_types=SimpleKineticMegacomplex,
    has_additional_penalty_function=lambda model: True,
    additional_penalty_function=additional_penalty_typecheck,
    has_matrix_constraints_function=lambda model: True,
    constrain_matrix_function=constrain_matrix_function_typecheck,
    retrieve_clp_function=retrieve_clp_typecheck,
    grouped=lambda model: model.is_grouped,
    index_dependent=lambda model: model.is_index_dependent,
)
class DecayModel(Model):
    additional_penalty_function_called = False
    constrain_matrix_function_called = False
    retrieve_clp_function_called = False
    is_grouped = False
    is_index_dependent = False


@model(
    "multi_channel",
    attributes={},
    dataset_type=GaussianShapeDecayDatasetDescriptor,
    model_dimension="c",
    global_matrix=calculate_spectral_gauss,
    global_dimension="e",
    megacomplex_types=SimpleKineticMegacomplex,
    grouped=lambda model: model.is_grouped,
    index_dependent=lambda model: model.is_index_dependent,
    has_additional_penalty_function=lambda model: True,
    additional_penalty_function=additional_penalty_typecheck,
)
class GaussianDecayModel(Model):
    additional_penalty_function_called = False
    constrain_matrix_function_called = False
    retrieve_clp_function_called = False
    is_grouped = False
    is_index_dependent = False


class OneCompartmentDecay:
    scale = 2
    wanted_parameters = ParameterGroup.from_list([101e-4])
    initial_parameters = ParameterGroup.from_list([100e-5, [scale, {"vary": False}]])

    e_axis = np.asarray([1.0])
    c_axis = np.arange(0, 150, 1.5)

    model_dict = {
        "megacomplex": {"m1": {}},
        "dataset": {
            "dataset1": {"initial_concentration": [], "megacomplex": ["m1"], "kinetic": ["1"]}
        },
    }
    sim_model = DecayModel.from_dict(model_dict)
    model_dict["dataset"]["dataset1"]["scale"] = "2"
    model = DecayModel.from_dict(model_dict)


class TwoCompartmentDecay:
    wanted_parameters = ParameterGroup.from_list([11e-4, 22e-5])
    initial_parameters = ParameterGroup.from_list([10e-4, 20e-5])

    e_axis = np.asarray([1.0])
    c_axis = np.arange(0, 150, 1.5)

    model = DecayModel.from_dict(
        {
            "megacomplex": {"m1": {}},
            "dataset": {
                "dataset1": {
                    "initial_concentration": [],
                    "megacomplex": ["m1"],
                    "kinetic": ["1", "2"],
                }
            },
        }
    )
    sim_model = model


class ThreeDatasetDecay:
    wanted_parameters = ParameterGroup.from_list([101e-4, 201e-3])
    initial_parameters = ParameterGroup.from_list([100e-5, 200e-3])

    e_axis = np.asarray([1.0])
    c_axis = np.arange(0, 150, 1.5)

    e_axis2 = np.asarray([1.0, 2.01])
    c_axis2 = np.arange(0, 100, 1.5)

    e_axis3 = np.asarray([0.99, 3.0])
    c_axis3 = np.arange(0, 150, 1.5)

    model_dict = {
        "megacomplex": {"m1": {}},
        "dataset": {
            "dataset1": {"initial_concentration": [], "megacomplex": ["m1"], "kinetic": ["1"]},
            "dataset2": {
                "initial_concentration": [],
                "megacomplex": ["m1"],
                "kinetic": ["1", "2"],
            },
            "dataset3": {"initial_concentration": [], "megacomplex": ["m1"], "kinetic": ["2"]},
        },
    }
    sim_model = DecayModel.from_dict(model_dict)
    model = sim_model


class MultichannelMulticomponentDecay:
    wanted_parameters = ParameterGroup.from_dict(
        {
            "k": [0.006, 0.003, 0.0003, 0.03],
            "loc": [
                ["1", 14705],
                ["2", 13513],
                ["3", 14492],
                ["4", 14388],
            ],
            "amp": [
                ["1", 1],
                ["2", 2],
                ["3", 5],
                ["4", 20],
            ],
            "del": [
                ["1", 400],
                ["2", 100],
                ["3", 300],
                ["4", 200],
            ],
        }
    )
    initial_parameters = ParameterGroup.from_dict({"k": [0.006, 0.003, 0.0003, 0.03]})

    e_axis = np.arange(12820, 15120, 50)
    c_axis = np.arange(0, 150, 1.5)

    sim_model = GaussianDecayModel.from_dict(
        {
            "compartment": ["s1", "s2", "s3", "s4"],
            "megacomplex": {"m1": {}},
            "dataset": {
                "dataset1": {
                    "initial_concentration": [],
                    "megacomplex": ["m1"],
                    "kinetic": ["k.1", "k.2", "k.3", "k.4"],
                    "location": ["loc.1", "loc.2", "loc.3", "loc.4"],
                    "delta": ["del.1", "del.2", "del.3", "del.4"],
                    "amplitude": ["amp.1", "amp.2", "amp.3", "amp.4"],
                }
            },
        }
    )
    model = GaussianDecayModel.from_dict(
        {
            "compartment": ["s1", "s2", "s3", "s4"],
            "megacomplex": {"m1": {}},
            "dataset": {
                "dataset1": {
                    "initial_concentration": [],
                    "megacomplex": ["m1"],
                    "kinetic": ["k.1", "k.2", "k.3", "k.4"],
                }
            },
        }
    )



"""Package with deprecation tests and helper functions."""
from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from typing import Any
    from typing import Callable
    from typing import Mapping
    from typing import Sequence


def deprecation_warning_on_call_test_helper(
    deprecated_callable: Callable[..., Any],
    *,
    raise_exception=False,
    args: Sequence[Any] = [],
    kwargs: Mapping[str, Any] = {},
) -> Any:
    """Helperfunction to quickly test that a deprecated class or function warns.

    By default this ignores error when calling the function/class,
    since those tests are only supposed to test the deprecation itself.

    However if the code is reimplemented it should be at least tested
    to return the proper type.


    Parameters
    ----------
    deprecated_callable : Callable[..., Any]
        Deprecated function or class.
    raise_exception : bool
        Whether or not to reraise an exception e.g. by calling with wrong args.
    args : Sequence[Any]
        Positional arguments for deprecated_callable.
    kwargs : Mapping[str, Any], optional
        Keyword arguments for deprecated_callable.

    Returns
    -------
    Any
        Return value of deprecated_callable

    Raises
    ------
    Exception
        Exception caused by deprecated_callable if raise_exception is True.
    """
    with pytest.warns(DeprecationWarning):
        try:
            return deprecated_callable(*args, **kwargs)
        except Exception:
            if raise_exception:
                raise



"""A base class for global analysis models."""
from __future__ import annotations

import copy
from typing import TYPE_CHECKING

from glotaran.deprecation import deprecate
from glotaran.parameter import ParameterGroup
from glotaran.utils.ipython import MarkdownStr

if TYPE_CHECKING:
    import numpy as np
    import xarray as xr


class Model:
    """A base class for global analysis models."""

    @classmethod
    def from_dict(cls, model_dict_ref: dict) -> Model:
        """Creates a model from a dictionary.

        Parameters
        ----------
        model_dict :
            Dictionary containing the model.
        """

        model = cls()

        model_dict = copy.deepcopy(model_dict_ref)

        # iterate over items
        for name, attribute in list(model_dict.items()):

            # we determine if the item is known by the model by looking for
            # a setter with same name.

            if hasattr(model, f"set_{name}"):

                # get the set function
                model_set = getattr(model, f"set_{name}")

                for label, item in attribute.items():
                    # we retrieve the actual class from the signature
                    item_cls = model_set.__func__.__annotations__["item"]

                    is_typed = hasattr(item_cls, "_glotaran_model_attribute_typed")

                    if isinstance(item, dict):
                        if is_typed:
                            if "type" not in item and item_cls.get_default_type() is None:
                                raise ValueError(f"Missing type for attribute '{name}'")
                            item_type = item.get("type", item_cls.get_default_type())

                            types = item_cls._glotaran_model_attribute_types
                            if item_type not in types:
                                raise ValueError(
                                    f"Unknown type '{item_type}' for attribute '{name}'"
                                )
                            item_cls = types[item_type]
                        item["label"] = label
                        model_set(label, item_cls.from_dict(item))
                    elif isinstance(item, list):
                        if is_typed:
                            if len(item) < 2 and len(item) != 1:
                                raise ValueError(f"Missing type for attribute '{name}'")
                            item_type = item[0]
                            types = item_cls._glotaran_model_attribute_types

                            if item_type not in types:
                                raise ValueError(
                                    f"Unknown type '{item_type}' for attribute '{name}'"
                                )
                            item_cls = types[item_type]
                        item = [label] + item
                        model_set(label, item_cls.from_list(item))
                del model_dict[name]

            elif hasattr(model, f"add_{name}"):

                # get the set function
                add = getattr(model, f"add_{name}")

                # we retrieve the actual class from the signature
                for item in attribute:
                    item_cls = add.__func__.__annotations__["item"]
                    is_typed = hasattr(item_cls, "_glotaran_model_attribute_typed")
                    if isinstance(item, dict):
                        if is_typed:
                            if "type" not in item:
                                raise ValueError(f"Missing type for attribute '{name}'")
                            item_type = item["type"]

                            if item_type not in item_cls._glotaran_model_attribute_types:
                                raise ValueError(
                                    f"Unknown type '{item_type}' for attribute '{name}'"
                                )
                            item_cls = item_cls._glotaran_model_attribute_types[item_type]
                        add(item_cls.from_dict(item))
                    elif isinstance(item, list):
                        if is_typed:
                            if len(item) < 2 and len(item) != 1:
                                raise ValueError(f"Missing type for attribute '{name}'")
                            item_type = (
                                item[1]
                                if len(item) != 1 and hasattr(item_cls, "label")
                                else item[0]
                            )

                            if item_type not in item_cls._glotaran_model_attribute_types:
                                raise ValueError(
                                    f"Unknown type '{item_type}' for attribute '{name}'"
                                )
                            item_cls = item_cls._glotaran_model_attribute_types[item_type]
                        add(item_cls.from_list(item))
                del model_dict[name]

        return model

    @property
    def model_type(self) -> str:
        """The type of the model as human readable string."""
        return self._model_type

    def problem_list(self, parameters: ParameterGroup = None) -> list[str]:
        """
        Returns a list with all problems in the model and missing parameters if specified.

        Parameters
        ----------

        parameter :
            The parameter to validate.
        """
        problems = []

        attrs = getattr(self, "_glotaran_model_attributes")
        for attr in attrs:
            attr = getattr(self, attr)
            if isinstance(attr, list):
                for item in attr:
                    problems += item.validate(self, parameters=parameters)
            else:
                for _, item in attr.items():
                    problems += item.validate(self, parameters=parameters)

        return problems

    @deprecate(
        deprecated_qual_name_usage="glotaran.model.base_model.Model.simulate",
        new_qual_name_usage="glotaran.analysis.simulation.simulate",
        to_be_removed_in_version="0.6.0",
        importable_indices=(2, 1),
    )
    def simulate(
        self,
        dataset: str,
        parameters: ParameterGroup,
        axes: dict[str, np.ndarray] = None,
        clp: np.ndarray | xr.DataArray = None,
        noise: bool = False,
        noise_std_dev: float = 1.0,
        noise_seed: int = None,
    ) -> xr.Dataset:
        """Simulates the model.

        Parameters
        ----------
        dataset :
            Label of the dataset to simulate.
        parameter :
            The parameters for the simulation.
        axes :
            A dictionary with axes for simulation.
        clp :
            Conditionally linear parameters. Used instead of `model.global_matrix` if provided.
        noise :
            If `True` noise is added to the simulated data.
        noise_std_dev :
            The standard deviation of the noise.
        noise_seed :
            Seed for the noise.
        """
        from glotaran.analysis.simulation import simulate

        return simulate(
            self,
            dataset,
            parameters,
            axes=axes,
            clp=clp,
            noise=noise,
            noise_std_dev=noise_std_dev,
            noise_seed=noise_seed,
        )

    def validate(self, parameters: ParameterGroup = None) -> str:
        """
        Returns a string listing all problems in the model and missing parameters if specified.

        Parameters
        ----------

        parameter :
            The parameter to validate.
        """
        result = ""

        problems = self.problem_list(parameters)
        if problems:
            result = f"Your model has {len(problems)} problems:\n"
            for p in problems:
                result += f"\n * {p}"
        else:
            result = "Your model is valid."
        return result

    def valid(self, parameters: ParameterGroup = None) -> bool:
        """Returns `True` if the number problems in the model is 0, else `False`

        Parameters
        ----------

        parameter :
            The parameter to validate.
        """
        return len(self.problem_list(parameters)) == 0

    def markdown(
        self,
        parameters: ParameterGroup = None,
        initial_parameters: ParameterGroup = None,
        base_heading_level: int = 1,
    ) -> MarkdownStr:
        """Formats the model as Markdown string.

        Parameters will be included if specified.

        Parameters
        ----------
        parameter: ParameterGroup
            Parameter to include.
        initial_parameters: ParameterGroup
            Initial values for the parameters.
        base_heading_level: int
            Base heading level of the markdown sections.

            E.g.:

            - If it is 1 the string will start with '# Model'.
            - If it is 3 the string will start with '### Model'.
        """
        base_heading = "#" * base_heading_level
        attrs = getattr(self, "_glotaran_model_attributes")
        string = f"{base_heading} Model\n\n"
        string += f"_Type_: {self.model_type}\n\n"

        for attr in attrs:
            child_attr = getattr(self, attr)
            if not child_attr:
                continue

            string += f"{base_heading}# {attr.replace('_', ' ').title()}\n\n"

            if isinstance(child_attr, dict):
                child_attr = child_attr.values()
            for item in child_attr:
                item_str = item.mprint(
                    parameters=parameters, initial_parameters=initial_parameters
                ).split("\n")
                string += f"* {item_str[0]}\n"
                for s in item_str[1:]:
                    string += f"  {s}\n"
            string += "\n"
        return MarkdownStr(string)

    def _repr_markdown_(self) -> str:
        """Special method used by ``ipython`` to render markdown."""
        return str(self.markdown(base_heading_level=3))

    def __str__(self):
        return str(self.markdown())



from __future__ import annotations

import dataclasses
import os
import pathlib
from typing import TYPE_CHECKING

import yaml

from glotaran.builtin.io.yml.sanatize import sanitize_yaml
from glotaran.io import ProjectIoInterface
from glotaran.io import load_dataset
from glotaran.io import load_model
from glotaran.io import load_parameters
from glotaran.io import register_project_io
from glotaran.io import save_dataset
from glotaran.io import save_parameters
from glotaran.model import get_model
from glotaran.parameter import ParameterGroup
from glotaran.project import SavingOptions
from glotaran.project import Scheme

if TYPE_CHECKING:
    from glotaran.model import Model
    from glotaran.project import Result


@register_project_io(["yml", "yaml", "yml_str"])
class YmlProjectIo(ProjectIoInterface):
    def load_model(self, file_name: str) -> Model:
        """parse_yaml_file reads the given file and parses its content as YML.

        Parameters
        ----------
        filename : str
            filename is the of the file to parse.

        Returns
        -------
        Model
            The content of the file as dictionary.
        """

        if self.format == "yml_str":
            spec = yaml.safe_load(file_name)

        else:
            with open(file_name) as f:
                spec = yaml.safe_load(f)

        spec = sanitize_yaml(spec)

        if "type" not in spec:
            raise Exception("Model type not defined")

        model_type = spec["type"]
        del spec["type"]

        model = get_model(model_type)
        return model.from_dict(spec)

    def load_parameters(self, file_name: str) -> ParameterGroup:

        if self.format == "yml_str":
            spec = yaml.safe_load(file_name)
        else:
            with open(file_name) as f:
                spec = yaml.safe_load(f)

        if isinstance(spec, list):
            return ParameterGroup.from_list(spec)
        else:
            return ParameterGroup.from_dict(spec)

    def load_scheme(self, file_name: str) -> Scheme:
        if self.format == "yml_str":
            yml = file_name
        else:
            try:
                with open(file_name) as f:
                    yml = f.read()
            except Exception as e:
                raise OSError(f"Error opening scheme: {e}")

        try:
            scheme = yaml.safe_load(yml)
        except Exception as e:
            raise ValueError(f"Error parsing scheme: {e}")

        if "model" not in scheme:
            raise ValueError("Model file not specified.")

        try:
            model = load_model(scheme["model"])
        except Exception as e:
            raise ValueError(f"Error loading model: {e}")

        if "parameters" not in scheme:
            raise ValueError("Parameters file not specified.")

        try:
            parameters = load_parameters(scheme["parameters"])
        except Exception as e:
            raise ValueError(f"Error loading parameters: {e}")

        if "data" not in scheme:
            raise ValueError("No data specified.")

        data = {}
        for label, path in scheme["data"].items():
            data_format = scheme.get("data_format", None)
            path = str(pathlib.Path(path).resolve())

            try:
                data[label] = load_dataset(path, format_name=data_format)
            except Exception as e:
                raise ValueError(f"Error loading dataset '{label}': {e}")

        optimization_method = scheme.get("optimization_method", "TrustRegionReflection")
        nnls = scheme.get("non-negative-least-squares", False)
        nfev = scheme.get("maximum-number-function-evaluations", None)
        ftol = scheme.get("ftol", 1e-8)
        gtol = scheme.get("gtol", 1e-8)
        xtol = scheme.get("xtol", 1e-8)
        group_tolerance = scheme.get("group_tolerance", 0.0)
        saving = SavingOptions(**scheme.get("saving", {}))
        return Scheme(
            model=model,
            parameters=parameters,
            data=data,
            non_negative_least_squares=nnls,
            maximum_number_function_evaluations=nfev,
            ftol=ftol,
            gtol=gtol,
            xtol=xtol,
            group_tolerance=group_tolerance,
            optimization_method=optimization_method,
            saving=saving,
        )

    def save_scheme(self, scheme: Scheme, file_name: str):
        _write_dict(file_name, dataclasses.asdict(scheme))

    def save_result(self, result: Result, result_path: str):
        options = result.scheme.saving

        if os.path.exists(result_path):
            raise FileExistsError(f"The path '{result_path}' is already existing.")

        os.makedirs(result_path)

        if options.report:
            md_path = os.path.join(result_path, "result.md")
            with open(md_path, "w") as f:
                f.write(str(result.markdown()))

        scheme_path = os.path.join(result_path, "scheme.yml")
        result_scheme = dataclasses.replace(result.scheme)
        result = dataclasses.replace(result)
        result.scheme = scheme_path

        parameters_format = options.parameter_format

        initial_parameters_path = os.path.join(
            result_path, f"initial_parameters.{parameters_format}"
        )
        save_parameters(result.initial_parameters, initial_parameters_path, parameters_format)
        result.initial_parameters = initial_parameters_path
        result_scheme.parameters = initial_parameters_path

        optimized_parameters_path = os.path.join(
            result_path, f"optimized_parameters.{parameters_format}"
        )
        save_parameters(result.optimized_parameters, optimized_parameters_path, parameters_format)
        result.optimized_parameters = optimized_parameters_path

        dataset_format = options.data_format
        for label, dataset in result.data.items():
            dataset_path = os.path.join(result_path, f"{label}.{dataset_format}")
            save_dataset(dataset, dataset_path, dataset_format, saving_options=options)
            result.data[label] = dataset_path
            result_scheme.data[label] = dataset_path

        result_file_path = os.path.join(result_path, "result.yml")
        _write_dict(result_file_path, dataclasses.asdict(result))
        result_scheme.result_path = result_file_path

        self.save_scheme(scheme=result_scheme, file_name=scheme_path)


def _write_dict(file_name: str, d: dict):
    with open(file_name, "w") as f:
        f.write(yaml.dump(d))



from typing import Dict
from typing import List
from typing import Tuple

import pytest
from IPython.core.formatters import format_display_data

from glotaran.model import Megacomplex
from glotaran.model import Model
from glotaran.model import model
from glotaran.model import model_attribute
from glotaran.parameter import Parameter
from glotaran.parameter import ParameterGroup


@model_attribute(
    properties={
        "param": Parameter,
        "megacomplex": str,
        "param_list": List[Parameter],
        "default_item": {"type": int, "default": 42},
        "complex": {"type": Dict[Tuple[str, str], Parameter]},
    },
)
class MockAttr:
    pass


@model_attribute()
class MockMegacomplex(Megacomplex):
    pass


@model_attribute()
class MockMegacomplex2(Megacomplex):
    pass


@model(
    "mock_model",
    attributes={"test": MockAttr},
    megacomplex_types={
        "mock_megacomplex": MockMegacomplex,
        "mock_megacomplex2": MockMegacomplex2,
    },
    model_dimension="model",
    global_dimension="global",
)
class MockModel(Model):
    pass


@pytest.fixture
def mock_model():
    d = {
        "megacomplex": {"m1": {}, "m2": ["mock_megacomplex2"]},
        "weights": [
            {
                "datasets": ["d1", "d2"],
                "global_interval": (1, 4),
                "model_interval": (2, 3),
                "value": 5.4,
            }
        ],
        "test": {
            "t1": {
                "param": "foo",
                "megacomplex": "m1",
                "param_list": ["bar", "baz"],
                "complex": {("s1", "s2"): "baz"},
            },
            "t2": ["baz", "m2", ["foo"], 7, {}],
        },
        "dataset": {
            "dataset1": {
                "megacomplex": ["m1", "m2"],
                "scale": "scale_1",
            },
            "dataset2": [["m2"], ["bar"], "scale_2"],
        },
    }
    return MockModel.from_dict(d)


@pytest.fixture
def model_error():
    d = {
        "megacomplex": {"m1": {}, "m2": {}},
        "test": {
            "t1": {
                "param": "fool",
                "megacomplex": "mX",
                "param_list": ["bar", "bay"],
                "complex": {("s1", "s3"): "boz"},
            },
        },
        "dataset": {
            "dataset1": {
                "megacomplex": ["N1", "N2"],
                "scale": "scale_1",
            },
            "dataset2": [["mrX"], ["bar"], "scale_3"],
        },
    }
    return MockModel.from_dict(d)


@pytest.fixture
def parameter():
    params = [1, 2, ["foo", 3], ["bar", 4], ["baz", 2], ["scale_1", 2], ["scale_2", 8], 4e2]
    return ParameterGroup.from_list(params)


def test_model_misc(mock_model: Model):
    assert mock_model.model_type == "mock_model"
    assert isinstance(mock_model.megacomplex["m1"], MockMegacomplex)
    assert isinstance(mock_model.megacomplex["m2"], MockMegacomplex2)


@pytest.mark.parametrize("attr", ["dataset", "megacomplex", "weights", "test"])
def test_model_attr(mock_model: Model, attr: str):
    assert hasattr(mock_model, attr)
    if attr != "weights":
        assert hasattr(mock_model, f"get_{attr}")
        assert hasattr(mock_model, f"set_{attr}")
    else:
        assert hasattr(mock_model, f"add_{attr}")


def test_model_validity(mock_model: Model, model_error: Model, parameter: ParameterGroup):
    print(mock_model.test["t1"])
    print(mock_model.problem_list())
    print(mock_model.problem_list(parameter))
    assert mock_model.valid()
    assert mock_model.valid(parameter)
    print(model_error.problem_list())
    print(model_error.problem_list(parameter))
    assert not model_error.valid()
    assert len(model_error.problem_list()) == 4
    assert not model_error.valid(parameter)
    assert len(model_error.problem_list(parameter)) == 8


def test_items(mock_model: Model):

    assert "m1" in mock_model.megacomplex
    assert "m2" in mock_model.megacomplex

    assert "t1" in mock_model.test
    t = mock_model.get_test("t1")
    assert t.param.full_label == "foo"
    assert t.megacomplex == "m1"
    assert [p.full_label for p in t.param_list] == ["bar", "baz"]
    assert t.default_item == 42
    assert ("s1", "s2") in t.complex
    assert t.complex[("s1", "s2")].full_label == "baz"
    assert "t2" in mock_model.test
    t = mock_model.get_test("t2")
    assert t.param.full_label == "baz"
    assert t.megacomplex == "m2"
    assert [p.full_label for p in t.param_list] == ["foo"]
    assert t.default_item == 7
    assert t.complex == {}

    assert "dataset1" in mock_model.dataset
    assert mock_model.get_dataset("dataset1").megacomplex == ["m1", "m2"]
    assert mock_model.get_dataset("dataset1").scale.full_label == "scale_1"

    assert "dataset2" in mock_model.dataset
    assert mock_model.get_dataset("dataset2").megacomplex == ["m2"]
    assert mock_model.get_dataset("dataset2").scale.full_label == "scale_2"

    assert len(mock_model.weights) == 1
    w = mock_model.weights[0]
    assert w.datasets == ["d1", "d2"]
    assert w.global_interval == (1, 4)
    assert w.model_interval == (2, 3)
    assert w.value == 5.4


def test_fill(mock_model: Model, parameter: ParameterGroup):
    dataset = mock_model.get_dataset("dataset1").fill(mock_model, parameter)
    assert [cmplx.label for cmplx in dataset.megacomplex] == ["m1", "m2"]
    assert dataset.scale == 2

    dataset = mock_model.get_dataset("dataset2").fill(mock_model, parameter)
    assert [cmplx.label for cmplx in dataset.megacomplex] == ["m2"]
    assert dataset.scale == 8

    t = mock_model.get_test("t1").fill(mock_model, parameter)
    assert t.param == 3
    assert t.megacomplex.label == "m1"
    assert t.param_list == [4, 2]
    assert t.default_item == 42
    assert t.complex == {("s1", "s2"): 2}
    t = mock_model.get_test("t2").fill(mock_model, parameter)
    assert t.param == 2
    assert t.megacomplex.label == "m2"
    assert t.param_list == [3]
    assert t.default_item == 7
    assert t.complex == {}


def test_model_markdown_base_heading_level(mock_model: Model):
    """base_heading_level applies to all sections."""
    assert mock_model.markdown().startswith("# Model")
    assert "## Test" in mock_model.markdown()
    assert mock_model.markdown(base_heading_level=3).startswith("### Model")
    assert "#### Test" in mock_model.markdown(base_heading_level=3)


def test_model_ipython_rendering(mock_model: Model):
    """Autorendering in ipython"""
    rendered_obj = format_display_data(mock_model)[0]

    assert "text/markdown" in rendered_obj
    assert rendered_obj["text/markdown"].startswith("### Model")

    rendered_markdown_return = format_display_data(mock_model.markdown())[0]

    assert "text/markdown" in rendered_markdown_return
    assert rendered_markdown_return["text/markdown"].startswith("# Model")



""" Kinetic Image Dataset Descriptor"""

from glotaran.model import DatasetDescriptor
from glotaran.model import model_attribute


@model_attribute(
    properties={
        "initial_concentration": {"type": str, "allow_none": True},
        "irf": {"type": str, "allow_none": True},
    }
)
class KineticImageDatasetDescriptor(DatasetDescriptor):
    pass



import numpy as np
import pytest
import xarray as xr

from glotaran.analysis.optimize import optimize
from glotaran.analysis.simulation import simulate
from glotaran.analysis.test.models import DecayModel
from glotaran.analysis.test.models import MultichannelMulticomponentDecay
from glotaran.analysis.test.models import OneCompartmentDecay
from glotaran.analysis.test.models import ThreeDatasetDecay
from glotaran.analysis.test.models import TwoCompartmentDecay
from glotaran.project import Scheme


@pytest.mark.parametrize("index_dependent", [True, False])
@pytest.mark.parametrize("grouped", [True, False])
@pytest.mark.parametrize("weight", [True, False])
@pytest.mark.parametrize(
    "method",
    [
        "TrustRegionReflection",
        "Dogbox",
        "Levenberg-Marquardt",
    ],
)
@pytest.mark.parametrize(
    "suite",
    # MultichannelMulticomponentDecay],
    [OneCompartmentDecay, TwoCompartmentDecay, ThreeDatasetDecay, MultichannelMulticomponentDecay],
)
def test_optimization(suite, index_dependent, grouped, weight, method):
    model = suite.model

    model.is_grouped = grouped
    model.is_index_dependent = index_dependent
    print("Grouped:", grouped)
    print("Index dependent:", index_dependent)

    assert model.grouped() == grouped
    assert model.index_dependent() == index_dependent

    sim_model = suite.sim_model
    sim_model.is_grouped = grouped
    sim_model.is_index_dependent = index_dependent

    print(model.validate())
    assert model.valid()

    print(sim_model.validate())
    assert sim_model.valid()

    wanted_parameters = suite.wanted_parameters
    print(wanted_parameters)
    print(sim_model.validate(wanted_parameters))
    assert sim_model.valid(wanted_parameters)

    initial_parameters = suite.initial_parameters
    print(initial_parameters)
    print(model.validate(initial_parameters))
    assert model.valid(initial_parameters)

    nr_datasets = 3 if issubclass(suite, ThreeDatasetDecay) else 1
    data = {}
    for i in range(nr_datasets):
        e_axis = getattr(suite, "e_axis" if i == 0 else f"e_axis{i+1}")
        c_axis = getattr(suite, "c_axis" if i == 0 else f"c_axis{i+1}")

        dataset = simulate(
            sim_model, f"dataset{i+1}", wanted_parameters, {"e": e_axis, "c": c_axis}
        )
        print(f"Dataset {i+1}")
        print("=============")
        print(dataset)

        if hasattr(suite, "scale"):
            dataset["data"] /= suite.scale

        if weight:
            dataset["weight"] = xr.DataArray(
                np.ones_like(dataset.data) * 0.5, coords=dataset.coords
            )

        assert dataset.data.shape == (c_axis.size, e_axis.size)

        data[f"dataset{i+1}"] = dataset

    scheme = Scheme(
        model=model,
        parameters=initial_parameters,
        data=data,
        maximum_number_function_evaluations=10,
        group_tolerance=0.1,
        optimization_method=method,
    )

    result = optimize(scheme)
    print(result.optimized_parameters)
    assert result.success
    optimized_scheme = result.get_scheme()
    assert result.optimized_parameters == optimized_scheme.parameters
    for dataset in optimized_scheme.data.values():
        assert "fitted_data" not in dataset
        if weight:
            assert "weight" in dataset
    for label, param in result.optimized_parameters.all():
        if param.vary:
            assert np.allclose(param.value, wanted_parameters.get(label).value, rtol=1e-1)

    for i, dataset in enumerate(data.values()):
        resultdata = result.data[f"dataset{i+1}"]
        print(f"Result Data {i+1}")
        print("=================")
        print(resultdata)
        assert "residual" in resultdata
        assert "residual_left_singular_vectors" in resultdata
        assert "residual_right_singular_vectors" in resultdata
        assert "residual_singular_values" in resultdata
        assert np.array_equal(dataset.c, resultdata.c)
        assert np.array_equal(dataset.e, resultdata.e)
        assert dataset.data.shape == resultdata.data.shape
        print(dataset.data[0, 0], resultdata.data[0, 0])
        assert np.allclose(dataset.data, resultdata.data)
        if weight:
            assert "weight" in resultdata
            assert "weighted_data" in resultdata
            assert np.allclose(resultdata.data, resultdata.weighted_data * 2)
            assert "weighted_residual" in resultdata
            assert "weighted_residual_left_singular_vectors" in resultdata
            assert "weighted_residual_right_singular_vectors" in resultdata
            assert "weighted_residual_singular_values" in resultdata

    assert callable(model.additional_penalty_function)
    assert model.additional_penalty_function_called

    if isinstance(model, DecayModel):
        assert callable(model.constrain_matrix_function)
        assert model.constrain_matrix_function_called
        assert callable(model.retrieve_clp_function)
        assert model.retrieve_clp_function_called
    else:
        assert not model.constrain_matrix_function_called
        assert not model.retrieve_clp_function_called



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from glotaran.deprecation import deprecate
from glotaran.io import load_scheme
from glotaran.utils.ipython import MarkdownStr

if TYPE_CHECKING:

    from typing import Literal

    import xarray as xr

    from glotaran.model import Model
    from glotaran.parameter import ParameterGroup

default_data_filters = {"minimal": ["fitted_data", "residual"], "full": None}


@dataclass
class SavingOptions:
    level: Literal["minimal", "full"] = "full"
    data_filter: list[str] | None = None
    data_format: str = "nc"
    parameter_format: str = "csv"
    report: bool = True


@dataclass
class Scheme:
    model: Model | str
    parameters: ParameterGroup | str
    data: dict[str, xr.DataArray | xr.Dataset | str]
    group_tolerance: float = 0.0
    non_negative_least_squares: bool = False
    maximum_number_function_evaluations: int = None
    ftol: float = 1e-8
    gtol: float = 1e-8
    xtol: float = 1e-8
    optimization_method: Literal[
        "TrustRegionReflection",
        "Dogbox",
        "Levenberg-Marquardt",
    ] = "TrustRegionReflection"
    saving: SavingOptions = SavingOptions()
    result_path: str | None = None

    def problem_list(self) -> list[str]:
        """Returns a list with all problems in the model and missing parameters."""
        return self.model.problem_list(self.parameters)

    def validate(self) -> str:
        """Returns a string listing all problems in the model and missing parameters."""
        return self.model.validate(self.parameters)

    def valid(self, parameters: ParameterGroup = None) -> bool:
        """Returns `True` if there are no problems with the model or the parameters,
        else `False`."""
        return self.model.valid(parameters)

    def markdown(self):
        """Formats the :class:`Scheme` as markdown string."""
        markdown_str = self.model.markdown(parameters=self.parameters)

        markdown_str += "\n\n"
        markdown_str += "__Scheme__\n\n"

        markdown_str += f"* *nnls*: {self.non_negative_least_squares}\n"
        markdown_str += f"* *nfev*: {self.maximum_number_function_evaluations}\n"
        markdown_str += f"* *group_tolerance*: {self.group_tolerance}\n"

        return MarkdownStr(markdown_str)

    def _repr_markdown_(self) -> str:
        """Special method used by ``ipython`` to render markdown."""
        return str(self.markdown())

    def __str__(self):
        """Representation used by print and str."""
        return str(self.markdown())

    @staticmethod
    @deprecate(
        deprecated_qual_name_usage="glotaran.project.scheme.Scheme.from_yaml_file(filename)",
        new_qual_name_usage="glotaran.io.load_scheme(filename)",
        to_be_removed_in_version="0.6.0",
        importable_indices=(2, 1),
    )
    def from_yaml_file(filename: str) -> Scheme:
        """Create :class:`Scheme` from specs in yaml file.

        Warning
        -------
        Deprecated use ``glotaran.io.load_scheme(filename)`` instead.

        Parameters
        ----------
        filename : str
            Path to the spec file.

        Returns
        -------
        Scheme
            Analysis schmeme
        """
        return load_scheme(filename)



import pytest
import xarray as xr
from IPython.core.formatters import format_display_data

from glotaran.io import load_scheme
from glotaran.project import Scheme


@pytest.fixture
def mock_scheme(tmpdir):

    model_path = tmpdir.join("model.yml")
    with open(model_path, "w") as f:
        model = "type: kinetic-spectrum\ndataset:\n  dataset1:\n    megacomplex: []"
        f.write(model)

    parameter_path = tmpdir.join("parameters.yml")
    with open(parameter_path, "w") as f:
        parameter = "[1.0, 67.0]"
        f.write(parameter)

    dataset_path = tmpdir.join("dataset.nc")
    xr.DataArray([[1, 2, 3]], coords=[("e", [1]), ("c", [1, 2, 3])]).to_dataset(
        name="data"
    ).to_netcdf(dataset_path)

    scheme = f"""
    model: {model_path}
    parameters: {parameter_path}
    non-negative-least-squares: True
    maximum-number-function-evaluations: 42
    data:
      dataset1: {dataset_path}

    saving:
        level: minimal
        data_filter: [a, b, c]
        data_format: csv
        parameter_format: yaml
        report: false
    """
    scheme_path = tmpdir.join("scheme.yml")
    with open(scheme_path, "w") as f:
        f.write(scheme)

    yield load_scheme(scheme_path)


def test_scheme(mock_scheme: Scheme):
    assert mock_scheme.model is not None
    assert mock_scheme.model.model_type == "kinetic-spectrum"

    assert mock_scheme.parameters is not None
    assert mock_scheme.parameters.get("1") == 1.0
    assert mock_scheme.parameters.get("2") == 67.0

    assert mock_scheme.non_negative_least_squares
    assert mock_scheme.maximum_number_function_evaluations == 42

    assert "dataset1" in mock_scheme.data
    assert mock_scheme.data["dataset1"].data.shape == (1, 3)

    assert mock_scheme.saving.level == "minimal"
    assert mock_scheme.saving.data_filter == ["a", "b", "c"]
    assert mock_scheme.saving.data_format == "csv"
    assert mock_scheme.saving.parameter_format == "yaml"
    assert not mock_scheme.saving.report


def test_scheme_ipython_rendering(mock_scheme: Scheme):
    """Autorendering in ipython"""

    rendered_obj = format_display_data(mock_scheme)[0]

    assert "text/markdown" in rendered_obj
    assert rendered_obj["text/markdown"].startswith("# Model")

    rendered_markdown_return = format_display_data(mock_scheme.markdown())[0]

    assert "text/markdown" in rendered_markdown_return
    assert rendered_markdown_return["text/markdown"].startswith("# Model")



from __future__ import annotations

from typing import Any

from glotaran.builtin.models.kinetic_image.irf import IrfMultiGaussian
from glotaran.parameter import Parameter

class IrfSpectralMultiGaussian(IrfMultiGaussian):
    @property
    def dispersion_center(self) -> Parameter: ...
    @property
    def center_dispersion(self) -> list[Parameter]: ...
    @property
    def width_dispersion(self) -> list[Parameter]: ...
    @property
    def model_dispersion_with_wavenumber(self) -> bool: ...
    def parameter(self, index: Any): ...
    def calculate_dispersion(self, axis: Any): ...

class IrfSpectralGaussian(IrfSpectralMultiGaussian):
    @property
    def center(self) -> Parameter: ...
    @property
    def width(self) -> Parameter: ...

class IrfGaussianCoherentArtifact(IrfSpectralGaussian):
    @property
    def coherent_artifact_order(self) -> int: ...
    @property
    def coherent_artifact_width(self) -> Parameter: ...
    def clp_labels(self): ...
    def calculate_coherent_artifact(self, axis: Any): ...



from __future__ import annotations

from typing import Any

from glotaran.model import model_attribute
from glotaran.model import model_attribute_typed
from glotaran.parameter import Parameter

class IrfMeasured: ...  # noqa: E701

class IrfMultiGaussian:
    @property
    def center(self) -> list[Parameter]: ...
    @property
    def width(self) -> list[Parameter]: ...
    @property
    def scale(self) -> list[Parameter]: ...
    @property
    def backsweep_period(self) -> Parameter: ...
    def parameter(self, index: Any): ...
    def calculate(self, index: Any, axis: Any): ...

class IrfGaussian(IrfMultiGaussian):
    @property
    def center(self) -> Parameter: ...
    @property
    def width(self) -> Parameter: ...

class Irf:
    @classmethod
    def add_type(cls, type_name: str, attribute_type: type) -> None: ...



"""The result class for global analysis."""
from __future__ import annotations

from dataclasses import dataclass

import numpy as np
import xarray as xr
from numpy.typing import ArrayLike
from tabulate import tabulate

from glotaran.deprecation import deprecate
from glotaran.io import save_result
from glotaran.model import Model
from glotaran.parameter import ParameterGroup
from glotaran.project.scheme import Scheme
from glotaran.utils.ipython import MarkdownStr


@dataclass
class Result:
    """The result of a global analysis"""

    additional_penalty: np.ndarray | None
    """A vector with the value for each additional penalty, or None"""
    cost: ArrayLike
    data: dict[str, xr.Dataset]
    """The resulting data as a dictionary of :xarraydoc:`Dataset`.

    Notes
    -----
    The actual content of the data depends on the actual model and can be found in the
    documentation for the model.
    """
    free_parameter_labels: list[str]
    """List of labels of the free parameters used in optimization."""
    number_of_function_evaluations: int
    """The number of function evaluations."""
    initial_parameters: ParameterGroup
    optimized_parameters: ParameterGroup
    """The optimized parameters, organized in a :class:`ParameterGroup`"""
    scheme: Scheme
    success: bool
    """Indicates if the optimization was successful."""
    termination_reason: str
    """The reason (message when) the optimizer terminated"""

    # The below can be none in case of unsuccessful optimization
    chi_square: float | None = None
    r"""The chi-square of the optimization.

    :math:`\chi^2 = \sum_i^N [{Residual}_i]^2`."""
    covariance_matrix: ArrayLike | None = None
    """Covariance matrix.

    The rows and columns are corresponding to :attr:`free_parameter_labels`."""
    degrees_of_freedom: int | None = None
    """Degrees of freedom in optimization :math:`N - N_{vars}`."""
    jacobian: ArrayLike | None = None
    """Modified Jacobian matrix at the solution

    See also: :func:`scipy.optimize.least_squares`
    """
    number_of_data_points: int | None = None
    """Number of data points :math:`N`."""
    number_of_jacobian_evaluations: int | None = None
    """The number of jacobian evaluations."""
    number_of_variables: int | None = None
    """Number of variables in optimization :math:`N_{vars}`"""
    optimality: float | None = None
    reduced_chi_square: float | None = None
    r"""The reduced chi-square of the optimization.

    :math:`\chi^2_{red}= {\chi^2} / {(N - N_{vars})}`.
    """
    root_mean_square_error: float | None = None
    r"""
    The root mean square error the optimization.

    :math:`rms = \sqrt{\chi^2_{red}}`
    """

    @property
    def model(self) -> Model:
        return self.scheme.model

    def get_scheme(self) -> Scheme:
        """Return a new scheme from the Result object with optimized parameters.

        Returns
        -------
        Scheme
            A new scheme with the parameters set to the optimized values.
            For the dataset weights the (precomputed) weights from the original scheme are used.
        """
        data = {}

        for label, dataset in self.data.items():
            data[label] = dataset.data.to_dataset(name="data")
            if "weight" in dataset:
                data[label]["weight"] = dataset.weight

        return Scheme(
            model=self.model,
            parameters=self.optimized_parameters,
            data=data,
            group_tolerance=self.scheme.group_tolerance,
            non_negative_least_squares=self.scheme.non_negative_least_squares,
            maximum_number_function_evaluations=self.scheme.maximum_number_function_evaluations,
            ftol=self.scheme.ftol,
            gtol=self.scheme.gtol,
            xtol=self.scheme.xtol,
            optimization_method=self.scheme.optimization_method,
        )

    def markdown(self, with_model: bool = True, base_heading_level: int = 1) -> MarkdownStr:
        """Formats the model as a markdown text.

        Parameters
        ----------
        with_model :
            If `True`, the model will be printed with initial and optimized parameters filled in.
        """

        general_table_rows = [
            ["Number of residual evaluation", self.number_of_function_evaluations],
            ["Number of variables", self.number_of_variables],
            ["Number of datapoints", self.number_of_data_points],
            ["Degrees of freedom", self.degrees_of_freedom],
            ["Chi Square", f"{self.chi_square or np.nan:.2e}"],
            ["Reduced Chi Square", f"{self.reduced_chi_square or np.nan:.2e}"],
            ["Root Mean Square Error (RMSE)", f"{self.root_mean_square_error or np.nan:.2e}"],
        ]
        if self.additional_penalty is not None:
            general_table_rows.append(["RMSE additional penalty", self.additional_penalty])

        result_table = tabulate(
            general_table_rows,
            headers=["Optimization Result", ""],
            tablefmt="github",
            disable_numparse=True,
        )
        if len(self.data) > 1:

            RMSE_rows = [
                [
                    f"{index}.{label}:",
                    dataset.weighted_root_mean_square_error,
                    dataset.root_mean_square_error,
                ]
                for index, (label, dataset) in enumerate(self.data.items(), start=1)
            ]

            RMSE_table = tabulate(
                RMSE_rows,
                headers=["RMSE (per dataset)", "weighted", "unweighted"],
                floatfmt=".2e",
                tablefmt="github",
            )

            result_table = f"{result_table}\n\n{RMSE_table}"

        if with_model:
            model_md = self.model.markdown(
                parameters=self.optimized_parameters,
                initial_parameters=self.initial_parameters,
                base_heading_level=base_heading_level,
            )
            result_table = f"{result_table}\n\n{model_md}"

        return MarkdownStr(result_table)

    def _repr_markdown_(self) -> str:
        """Special method used by ``ipython`` to render markdown."""
        return str(self.markdown(base_heading_level=3))

    def __str__(self):
        return str(self.markdown(with_model=False))

    @deprecate(
        deprecated_qual_name_usage="glotaran.project.result.Result.save(result_path)",
        new_qual_name_usage=(
            "glotaran.io.save_result("
            "result=result, result_path=result_path, "
            'format_name="legacy", allow_overwrite=True'
            ")"
        ),
        to_be_removed_in_version="0.6.0",
        importable_indices=(2, 1),
    )
    def save(self, path: str) -> list[str]:
        """Saves the result to given folder.

        Warning
        -------
        Deprecated use ``save_result(result_path=result_path, result=result,
        format_name="legacy", allow_overwrite=True)`` instead.


        Returns a list with paths of all saved items.
        The following files are saved:

        * `result.md`: The result with the model formatted as markdown text.

        * `optimized_parameters.csv`: The optimized parameter as csv file.

        * `{dataset_label}.nc`: The result data for each dataset as NetCDF file.

        Parameters
        ----------
        path :
            The path to the folder in which to save the result.
        """
        save_result(result_path=path, result=self, format_name="legacy", allow_overwrite=True)

    @deprecate(
        deprecated_qual_name_usage="glotaran.project.result.Result.get_dataset(dataset_label)",
        new_qual_name_usage=("glotaran.project.result.Result.data[dataset_label]"),
        to_be_removed_in_version="0.6.0",
        importable_indices=(2, 2),
    )
    def get_dataset(self, dataset_label: str) -> xr.Dataset:
        """Returns the result dataset for the given dataset label.

        Warning
        -------
        Deprecated use ``glotaran.project.result.Result.data[dataset_label]``
        instead.


        Parameters
        ----------
        dataset_label :
            The label of the dataset.
        """
        try:
            return self.data[dataset_label]
        except KeyError:
            raise ValueError(f"Unknown dataset '{dataset_label}'")



from __future__ import annotations

import pytest
from IPython.core.formatters import format_display_data

from glotaran.analysis.optimize import optimize
from glotaran.analysis.simulation import simulate
from glotaran.analysis.test.models import ThreeDatasetDecay as suite
from glotaran.project import Scheme
from glotaran.project.result import Result


@pytest.fixture(scope="session")
def dummy_result():
    """Dummy result for testing."""

    model = suite.model

    model.is_grouped = False
    model.is_index_dependent = False

    wanted_parameters = suite.wanted_parameters
    data = {}
    for i in range(3):
        e_axis = getattr(suite, "e_axis" if i == 0 else f"e_axis{i+1}")
        c_axis = getattr(suite, "c_axis" if i == 0 else f"c_axis{i+1}")

        data[f"dataset{i+1}"] = simulate(
            suite.sim_model, f"dataset{i+1}", wanted_parameters, {"e": e_axis, "c": c_axis}
        )
    scheme = Scheme(
        model=suite.model,
        parameters=suite.initial_parameters,
        data=data,
        maximum_number_function_evaluations=1,
    )

    yield optimize(scheme)


def test_result_ipython_rendering(dummy_result: Result):
    """Autorendering in ipython"""

    rendered_obj = format_display_data(dummy_result)[0]

    assert "text/markdown" in rendered_obj
    assert rendered_obj["text/markdown"].startswith("| Optimization Result")

    rendered_markdown_return = format_display_data(dummy_result.markdown())[0]

    assert "text/markdown" in rendered_markdown_return
    assert rendered_markdown_return["text/markdown"].startswith("| Optimization Result")



from __future__ import annotations

from glotaran.model import DatasetDescriptor
from glotaran.model import model_attribute
from glotaran.parameter import Parameter

class InitialConcentration:
    @property
    def compartments(self) -> list[str]: ...
    @property
    def parameters(self) -> list[Parameter]: ...
    @property
    def exclude_from_normalize(self) -> list[Parameter]: ...
    def normalized(self, dataset: DatasetDescriptor) -> InitialConcentration: ...
